from os import path, getcwd, mkdir
from collections import defaultdict
from jinja2 import Template
from jinja2.filters import FILTERS
from shutil import which, copy2, SameFileError
import logging
from cmath import sqrt
from subprocess import Popen, PIPE, STDOUT, TimeoutExpired
from anyBSM.utils import fieldtypes, module_dir
from anyBSM.latex.printer import latex, texname, line, particle_name
from sympy import sympify, SympifyError # TODO avoid imports everywhere
import re

logger = logging.getLogger('anyBSM.latex.drawer')

__doc__ = """
# LaTeX Output
Generates LaTeX summary of all calculated diagrams using [TikZ-Feynman](https://www.sciencedirect.com/science/article/abs/pii/S0010465516302521)

## Example
from anyBSM import anyBSM
from anyBSM.latex.drawer import DrawProcess

model = anyBSM('THDMII')
h=model.SM_particles['Higgs-Boson']

# calculated tadpole and draws result
tad=model.process(h, draw=True)

# the following is equivalent to above but gives you access to the drawer
drawer = DrawProcess(model, externalfields=[h])
model.process(h, draw=drawer)
# here you could calculate/draw more diagrams
# finalize/compile the latex output
drawer.write(result=tad)

"""

drawer_path = path.join(module_dir, 'latex')

FILTERS['line'] = line
FILTERS['texname'] = texname

def printparameters(parameters):
    string = ''
    r = 7 # number of digits
    ncol = 5 # number of columns
    colsep = 'l'.join(['' for i in range(ncol+1)])
    nparas = len(parameters)
    string += f'\\begin{{longtable}}{{{colsep}}}\n'
    for i,p in enumerate(sorted(parameters, key=lambda x: x.name.lower())):
        num = p.nvalue
        try:
            if abs(num.imag) > r+2:
                num = str(round(num.real, r) + round(num.imag, r) * 1j)
            else:
                num = str(round(num.real, r))
            num = latex(sympify(num))
        except AttributeError:
            logger.warning(f'cannot latexify parameter value {p.name}={p.nvalue}')
            num = '\\text{NaN}'
        if p.name != 'ZERO':
            string += f'${p.texname} = {num}$'
            if (i+1) % ncol == 0:
                string += ', \\\\\n'
            elif (i+1) != nparas:
                string += ', & '
    string += '\\end{longtable}'
    return string


class DrawProcess():
    r""" Draw a list of Feynman diagrams along with their results.
    Upon initialization the header of the $\LaTeX$ file is written to the
    model directory. Each call of `DrawProcess.add_diagram(topo,
    particles, result)` registers a new diagram
    of topology type `topo` (c.f. `anyBSM.anyBSM.topolgies`) and internal particles `particles` as well
    as its analytical/numerical `result` to be written to the file.
    Calling `DrawProcess.write()` writes the individual results to the
    file as well as generates an overview which lists all insertions.
    """
    model: object
    """ The `anyBSM.anyBSM()` model object """
    externalfields: tuple
    """ List of external fields of the amplitude (`anyBSM.ufo.object_library.Particle` objects) """
    simplify: bool
    """ whether to simplify analytic results """
    cutoff: object
    """ Whether to place a `cutoff` (numerical value) on which diagrams to draw. `False`: no cutoff. """
    def __init__(self,model, externalfields, momenta = [], simplify = False, cutoff = False, process_options = ''):
        self.file = path.join(model.modeldir,  'latex', model.name + '_' + ''.join(f.name for f in externalfields) + f'_{model.evaluation}_results.tex')
        if not path.exists(path.dirname(self.file)):
            mkdir(path.dirname(self.file))
        self.model = model
        self.latex = self.model.sympify_parameters()
        self.ext = externalfields
        self.mom = momenta
        self.simplify = simplify
        self.cutoff = cutoff if cutoff is not True else False
        self.next = len(externalfields)
        self.insertions = defaultdict(dict)
        self.diagrams = defaultdict(dict)
        options = re.sub(r': ([^;]*);',r': \\texttt{\1};',process_options)
        options = f'\\item Process function called with options:\\\\ {options}'
        if momenta:
            options += f'\\item Momenta: \\texttt{{{momenta}}}'
        options += f'\\item Renormalisation scheme (see schemes.yml): \\texttt{{{model.scheme_name}}}.'
        options += f'\\item Cut-off set to: ${self._unit(cutoff)}$ (plotting only diagrams with |result|$>{self._unit(cutoff)}$. Numerical results include all selected diagrams).' if self.cutoff else ''

        with open(self.file, 'w') as f:
            f.write(f"""
% MANUAL COMPILATION:
% lualatex --halt-on-error {path.basename(self.file)}
\\documentclass[11pt]{{article}}
\\usepackage[margin=2.5cm]{{geometry}}
\\usepackage{{amsmath}}
\\usepackage{{breqn}}
\\usepackage{{array}}
\\usepackage{{hyperref}}
\\usepackage{{ltablex}}
\\usepackage{{amssymb}}
\\usepackage{{tikz}}
\\usepackage[compat=1.1.0]{{{drawer_path}/tikzfeynman/tikzfeynman}}
\\usepackage{{array}}
\\usepackage{{longtable}}
\\usepackage{{units}}
\\hypersetup{{
    colorlinks=true,
    linkcolor=blue,
    urlcolor=cyan,
    }}
\\begin{{document}}
\\newcolumntype{{L}}[1]{{>{{\\raggedleft\\arraybackslash}}m{{#1}}}}
\\title{{
\\vspace*{{-3.5cm}}\\hspace{{-16cm}}\\includegraphics[width=0.2\\textwidth]{{{module_dir}/logos/anyH3_logo.pdf}}\\\\[-2cm]
\\href{{https://anybsm.gitlab.io}}{{anyBSM}} - Model: {model.name} (Amplitude: ${'-'.join([particle_name(f) for f in self.ext])}$)}}
\\author{{Henning Bahl, Johannes Braathen, Martin Gabelmann, Georg Weiglein}}
\\maketitle
\\noindent
\\section{{Setup}}
\\begin{{itemize}}
{options}
\\end{{itemize}}
% \\tableofcontents % uncomment and compile twice for TOC
% \\pagebreak
""")

    def _unit(self,num = ''):
        """ returns the unit (in GeV) of the amplitude in LaTeX-form."""
        u = 4 - self.next
        u = '' if u == 1 else f'^{u}'
        if num:
            return f'\\unit[{num}]{{\\text{{GeV}}{u}}}'
        return f'\\text{{GeV}}{u}'

    def add_diagram(self, topo, particles, result, externalfields=[], momenta = []):
        """ Register a diagram to draw.
            * `topo`: string that represents a topology defined in
            anyBSM.topologies
            * `particles`: list of
            `anyBSM.ufo.object_library.Particle()`
            * `result` numerical or analytical result of the diagram
            (number- or string-type)
        """
        if self.model.evaluation == 'numerical' and (self.cutoff and abs(result) < float(self.cutoff)):
            return
        inttype = fieldtypes(particles)
        if topo not in skeletons[self.next]:
            return
        extlabels = self.ext if not externalfields else externalfields
        if inttype not in self.insertions[topo]:
            self.insertions[topo][inttype] = [skeletons[self.next][topo].render(
                    external=extlabels,
                    momenta = [],
                    internal=particles,
                    generic=True
                    )]
        insert = '\\{' + ','.join([particle_name(f) for f in particles]) + '\\}'
        if inttype and insert not in self.insertions[topo][inttype]:
            self.insertions[topo][inttype].append(insert)

        if inttype not in self.diagrams[topo]:
            self.diagrams[topo][inttype] = []
        norm = '64*pi**2'
        normtex = ' $64\\pi^2\\times$ '
        if 'Tree' in topo or self.model.evaluation == 'numerical':
            norm = '1'
            normtex = ''

        result = '0' if result == '' else result
        if self.model.evaluation != 'numerical':
            result = self.model.sympify(f'{norm}*({result}+0)')
            result = self.latex(result, symbol_names = self.model.texnames)
            mom = [f', {self.latex(self.model.sympify(m))}' for m in momenta]
        else:
            mom = [f'^{{p^2=({round(sqrt(m).real)})^2}}' for m in momenta]
        self.diagrams[topo][inttype].append([
            normtex,
            skeletons[self.next][topo].render(
                external=extlabels,
                momenta = mom,
                internal=particles,
                generic=False),
            ' = & \\begin{dmath*}\n',
            result,
            '\n\\end{dmath*}\n \\\\'
        ])

    def write(self, result={}):
        """ Writes all registered diagrams to the latex file and try to
        compile it.
          * `result`: the result of `anyBSM.process()` i.e. a
          dictionary separating the summed results of the individial
          topologies.

        At the beginning of the file, an overview with all insertions
        for the various topologies is generated. If the numerical
        evaluation mode is used and `result` is given, an overview of the different
        contributions and their weights is listed.
        """
        f = open(self.file, 'a')
        f.write('\n\n\\section{Overview}\n')
        for topo in self.insertions:
            f.write(f'\\subsection{{Topology: {topo}}}\n\n')
            f.write('\\begin{tabularx}{\\textwidth}{@{}rlX@{}}')
            for inttype,ins in self.insertions[topo].items():
                f.write(ins[0])
                if inttype:
                    f.write('& $\\{' + ','.join(f for f in inttype) + '\\} =$ &')
                    f.write('$' + '$, $'.join(ins[1:min(len(ins), 50)]) + '$')
                    if len(ins) > 50:
                        f.write(f' (+{len(ins)-50} more insertions) \\\\ \n')
                    else:
                        f.write('\\\\ \n')
                else:
                    f.write('& & \\\\\n')
            f.write('\\end{tabularx}\n')
        if self.model.evaluation == 'numerical' and result:
            f.write('\n\\section{Numerical input values}\n')
            if len(self.ext) > 1:
                f.write('External momenta:\\\\\n')
                for i,m in enumerate(self.mom):
                    f.write(f'$\\sqrt{{p_{i+1}^2}} = \\sqrt{{p^2_{{{particle_name(self.ext[i])}}}}} = {sqrt(m).real}\\, \\text{{GeV}}$ ')
                    if len(self.ext) == 2:
                        continue
                    if i+1 < len(self.mom):
                        f.write(',\\\\')
            f.write('\n\\subsection{Input (external) parameters}\n')
            f.write(printparameters(self.model.external_parameters))
            f.write('\n\\subsection{Derived (internal) parameters}\n')
            f.write(printparameters(self.model.internal_parameters))
            result = dict(result)
            f.write('\n\\section{Summary of individual results}\n')
            for k,v in result.items():
                result[k] = eval(str(v))
            total = sum(eval(str(r)) for r in result.values())
            f.write('\\begin{align*}')
            f.write('\\textbf{Topology} & & \\textbf{contribution in ' + self._unit() + '} & & \\textbf{in \\%} \\\\ \\cline{1-5} \n')
            f.write(f'\\text{{Sum of all contributions}} & & {total} & & \\textbf{{(100\\%}})\\\\ \n')
            for k,v in result.items():
                f.write(f'\\text{{{k}}} & & {result[k]} & & \\textbf{{({round(100*result[k].real/total.real,3)}\\%}})\\\\ \n')
            f.write('\\end{align*}\n')
            f.write('\n\\textit{Relative (\\%) values are calculated for real parts only.}\\\\ \n')
            if len(self.ext) == 3:
                f.write("""
                        \\begin{itemize}
                        \\item ThreePointTree corresponds to the tree-level result of the three-point function,
                        \\item the sum of ThreePointB and ThreePointC corresponds to the genuine 1-loop result,
                        \\item the sum of ThreePointTA and ThreePointWFRT corresponds to the 1-loop tadpole contribution,
                        \\item the sum of ThreePointWFRA and ThreePointWFRB corresponds to the 1-loop external leg contribution,
                        \\item ThreePointCT is the sum of all (mass, VEV and  custom) counter-term contributions to the tree-point function.
                        \\end{itemize}\n\n""")
            f.write('See appendix A.2 of \\href{https://arxiv.org/abs/2305.03015}{2305.03015} for detailed explanations of the different topologies')

        f.write('\n\\section{Individual results}\n')
        for topo in self.diagrams:
            f.write(f'\\subsection{{Topology: {topo}}}\n')
            for inttype, diags in self.diagrams[topo].items():
                f.write(f'\\subsubsection{{{inttype}}}\n')
                f.write('\\begin{longtable}{rL{0.9\\textwidth}}')
                for diag in diags:
                    if self.model.evaluation == 'numerical' and abs(total) > 0:
                        try:
                            diag[3] = self._unit(latex(sympify(diag[3]))) + ' \\quad \\textbf{(' + str(round(diag[3].real*100/total.real,3)) + '\\%)}'
                        except SympifyError:
                            diag[3] = str(diag[3])
                    else:
                        diag[3] = str(diag[3])
                    f.write(''.join(diag) + '\n')
                f.write("\n\\end{longtable}\n")
        f.write("\n\\vfill\n\\textit{This document was compiled with the help of \\href{https://www.sciencedirect.com/science/article/abs/pii/S0010465516302521}{TikZ-Feynman}}")
        f.write("\\end{document}")
        f.close()
        lualatex = which('lualatex') or which('pdflatex')
        if not lualatex:
            logger.error('Cannot find any lualatex/pdflatex installation on your system.')
            return
        print(f"Compiling LaTeX ({self.file})...")
        proc = Popen(
                [lualatex, '--halt-on-error', self.file],
                stderr=STDOUT,
                stdout=PIPE,
                cwd=path.dirname(self.file)
        )
        try:
            stdout, stderr = proc.communicate(timeout=600)
            rc = proc.poll()
            fname = path.splitext(self.file)[0] + '.pdf'
            if rc != 0 or not path.exists(fname):
                logger.error('LaTeX compilation failed:')
                return
            print("done.")
            newpath = path.abspath(path.join(getcwd(), path.basename(fname)))
            try:
                copy2(fname, newpath)
            except SameFileError:
                pass
            print(f"Open the pdf file at: {newpath}")
        except TimeoutExpired:
            logger.error(f"Compiling LaTeX failed (timeout). Please try manually compiling {self.file}")
            proc.kill()


skeletons = defaultdict(dict)
""" skeletons[number_external_legs][topology_name] (identical to anyBSM.topologies.TOPOLOGIES)
contains Template expressions for Feynman diagrams of the different
topologies.
"""

skeletons[1]['OnePointA'] = Template(r"""
\begin{tikzpicture}[baseline=0]
    \begin{feynman}
        \vertex (in){${{ external[0]|texname(generic, True) }}$};
        \vertex[right=1.0cm of in](a);
        \vertex[right=1.5cm of a](b);
        \vertex[right=0.8cm of a](label) {${{ internal[0]|texname(generic) }}$};
        \vertex[right=0.6cm of b](tmo) {};
        \diagram*{(in) -- [{{external[0]|line}}] (a), (a) --[{{internal[0]|line}}, half right] (b) -- [{{internal[0]|line}}, half right] (a)};
    \end{feynman}
\end{tikzpicture}

""")

skeletons[2]['TwoPointA'] = Template(r"""
\begin{tikzpicture}[baseline=0]
    \begin{feynman}
        \vertex (in){${{ external[0]|texname(generic, True) }}$};
        \vertex[right=1.5cm of in](a);
        \vertex[above=1.5cm of a](b);
        \vertex[right=1.1cm of a](out) {${{ external[1]|texname(generic, True) }}$};
        \vertex[above=0.8cm of a](label) {${{ internal[0]|texname(generic) }}$};
        \diagram*{(in) -- [{{external[0]|line}}] (a), (a) --[{{external[1]|line}}] (out), (a) -- [{{internal[0]|line}}, half right] (b) -- [{{internal[0]|line}}, half right] (a)};
    \end{feynman}
\end{tikzpicture}

""")

skeletons[2]['TwoPointB'] = Template(r"""
\begin{tikzpicture}[baseline=0]
    \begin{feynman}
        \vertex (in) {${{ external[0]|texname(generic,True) }}$};
        \vertex[right=1cm of in](a);
        \vertex[right=1.5cm of a](b);
        \vertex[right=0.6cm of b](out) {${{ external[1]|texname(generic, True) }}$};
        \diagram*{(in)  -- [{{external[0]|line}}] (a), (b) -- [{{external[1]|line}}] (out),
                  (a)  -- [{{internal[0]|line}}, edge label=${{internal[0]|texname(generic)}}$,half left,looseness=1.59] (b),
                  (b)  -- [{{internal[1]|line}}, edge label=${{internal[1]|texname(generic)}}$,half left,looseness=1.59] (a)
                 };
    \end{feynman}
\end{tikzpicture}

""")

skeletons[2]['TwoPointTA'] = Template(r"""
\begin{tikzpicture}[baseline=0]
    \begin{feynman}
        \vertex (in){${{ external[0]|texname(generic, True) }}$};
        \vertex[right=1.5cm of in](a);
        \vertex[above=1cm of a](b);
        \vertex[right=1.1cm of a](out) {${{ external[1]|texname(generic, True) }}$};
        \vertex[above=0.8cm of a](label) {};
        \diagram*{(in) -- [{{external[0]|line}}] (a),
                  (a) --[{{external[1]|line}}] (out),
                  (a) -- [{{internal[0]|line}}, edge label=${{ internal[0]|texname(generic) }}$] (b)};
        \draw[{{internal[1]|line}}] (b) arc [start angle=-90, end angle=270, radius=0.5cm] node[label=${{internal[1]|texname(generic)}}$] {};
    \end{feynman}
\end{tikzpicture}

""")

skeletons[3]['ThreePointTree'] = Template(r"""
\begin{tikzpicture}[baseline=0]
    \begin{feynman}
        \vertex (X);
        \vertex[left=1cm of X] (e1) {${{ external[0]|texname(generic,True) }}$};
        \vertex[right=1cm of X] (tmp);
        \vertex[above=1cm of tmp] (e2) {${{ external[1]|texname(generic,True) }}$};
        \vertex[below=1cm of tmp] (e3) {${{ external[2]|texname(generic,True) }}$};
        \diagram*{(e1)  -- [{{external[0]|line}}] (X),
                  (e2)  -- [{{external[1]|line}}] (X),
                  (e3)  -- [{{external[2]|line}}] (X)};
    \end{feynman}
\end{tikzpicture}

""")

skeletons[3]['ThreePointB'] = Template(r"""
\begin{tikzpicture}[baseline=0]
    \begin{feynman}
        \vertex (in) {${{ external[0]|texname(generic,True) }} {{momenta[0]}}$ };
        \vertex[right=1cm of in](a);
        \vertex[right=1.5cm of a](b);
        \vertex[right=1cm of b](tmp);
        \vertex[above=0.56cm of tmp](out1) {${{ external[1]|texname(generic, True) }} {{momenta[1]}} $};
        \vertex[below=0.56cm of tmp](out2) {${{ external[2]|texname(generic, True) }} {{momenta[2]}} $};
        \diagram*{(in)  -- [{{external[0]|line}}] (a),
                  (out1) -- [{{external[1]|line}}] (b),
                  (out2) -- [{{external[2]|line}}] (b),
                  (a)  -- [{{internal[0]|line}}, edge label=${{internal[0]|texname(generic)}}$,half left,looseness=1.59] (b),
                  (b)  -- [{{internal[1]|line}}, edge label=${{internal[1]|texname(generic)}}$,half left,looseness=1.59] (a)
                 };
    \end{feynman}
\end{tikzpicture}

""")

skeletons[3]['ThreePointC'] = Template(r"""
\begin{tikzpicture}[baseline=(a)]
    \begin{feynman}
      \vertex at (0,1.75) (a) {${{ external[0]|texname(generic,True) }}$};
      \vertex at (3.5,0) (b) {${{ external[2]|texname(generic,True) }}$};
      \vertex at (3.5,3.5) (c) {${{ external[1]|texname(generic,True) }}$};

      \vertex at (1.16,1.75) (a1);
      \vertex at (2.62,0.58) (b1);
      \vertex at (2.62,2.917) (c1);

      \diagram* {
      (a) -- [{{external[0]|line}}] (a1),
      (b) -- [{{external[2]|line}}] (b1),
      (c) -- [{{external[1]|line}}] (c1),
      (a1) -- [{{internal[0]|line}}, edge label=${{internal[0]|texname(generic)}}$] (c1) -- [{{internal[1]|line}}, edge label=${{internal[2]|texname(generic)}}$] (b1) -- [{{internal[2]|line}}, edge label=${{internal[1]|texname(generic)}}$] (a1)
      };

    \end{feynman}
\end{tikzpicture}

""")

skeletons[3]['ThreePointTA'] = Template(r"""
\begin{tikzpicture}[baseline=(X.base)]
    \begin{feynman}
        \vertex (X);
        \vertex[left=1cm of X] (t1);
        \vertex[right=0.5cm of X] (t2);
        \vertex[right=1cm of X] (t3);
        \vertex[above=1cm of t1] (e1) {${{ external[0]|texname(generic,True) }}$};
        \vertex[below=1cm of t1] (e2) {${{ external[1]|texname(generic,True) }}$};
        \vertex[below=1cm of t3] (e3) {${{ external[2]|texname(generic,True) }}$};
        \vertex[above=0.5cm of t2] (x);
        \diagram*{(e1)  -- [{{external[0]|line}}] (X),
                  (e2)  -- [{{external[1]|line}}] (X),
                  (e3)  -- [{{external[2]|line}}] (X),
                  (X)  -- [{{internal[0]|line}}, edge label=${{internal[0]|texname(generic)}}$] (x)};
        \draw[{{internal[1]|line}}] (x) arc [start angle=-135, end angle=225, radius=0.55cm] node[label=20:${{internal[1]|texname(generic)}}$] {};
    \end{feynman}
\end{tikzpicture}

""")

skeletons[3]['ThreePointWFRA'] = Template(r"""
\begin{tikzpicture}[baseline=(X.base)]
    \begin{feynman}
        \vertex (X);
        \vertex[left=1cm of X] (e1) {${{ external[0]|texname(generic,True) }} {{momenta[0]}}$};
        \vertex[right=1cm of X] (tmp);
        \vertex[right=0.5cm of X] (tmp0);
        \vertex[above=0.5cm of tmp0] (x);
        \vertex[above=1cm of tmp] (e2) {${{ external[2]|texname(generic,True) }} {{momenta[2]}}$};
        \vertex[below=1cm of tmp] (e3) {${{ external[1]|texname(generic,True) }} {{momenta[1]}}$};
        \diagram*{(e1)  -- [{{external[0]|line}}] (X),
                  (e3)  -- [{{external[1]|line}}] (X),
                  (X)  -- [{{internal[0]|line}}, edge label'=${{internal[0]|texname(generic)}}$] (x),
                  (e2)  -- [{{external[2]|line}}] (x)};
        \draw[{{internal[1]|line}}] (x) arc [start angle=-45, end angle=315, radius=0.4cm] node[label={[xshift=-0.5cm,yshift=0.5cm]${{internal[1]|texname(generic)}}$}] {};
    \end{feynman}
\end{tikzpicture}

""")

skeletons[3]['ThreePointWFRB'] = Template(r"""
\begin{tikzpicture}[baseline=(X.base)]
    \begin{feynman}
        \vertex (X);
        \vertex[left=1cm of X] (e1) {${{ external[0]|texname(generic,True) }} {{momenta[0]}}$};
        \vertex[right=1.5cm of X] (tmp);
        \vertex[right=0.5cm of X] (tmp0);
        \vertex[right=1cm of X] (tmp1);
        \vertex[above=0.5cm of tmp0] (x1);
        \vertex[above=1cm of tmp1] (x2);
        \vertex[above=1.3cm of tmp] (e2) {${{ external[2]|texname(generic,True) }} {{momenta[2]}}$};
        \vertex[below=1.3cm of tmp] (e3) {${{ external[1]|texname(generic,True) }} {{momenta[1]}}$};
        \diagram*{(e1)  -- [{{external[0]|line}}] (X),
                  (e3)  -- [{{external[1]|line}}] (X),
                  (X)  -- [{{internal[0]|line}}, edge label=${{internal[0]|texname(generic)}}$] (x1),
                  (x1)  -- [{{internal[1]|line}}, edge label=${{internal[1]|texname(generic)}}$, half left] (x2),
                  (x2)  -- [{{internal[2]|line}}, edge label=${{internal[2]|texname(generic)}}$, half left] (x1),
                  (e2)  -- [{{external[2]|line}}] (x2)};
    \end{feynman}
\end{tikzpicture}

""")
skeletons[3]['ThreePointWFRT'] = Template(r"""
\begin{tikzpicture}[baseline=(X.base)]
    \begin{feynman}
        \vertex (X);
        \vertex[left=1cm of X] (e1) {${{ external[0]|texname(generic,True) }} {{momenta[0]}}$};
        \vertex[right=1cm of X] (tmp);
        \vertex[right=0.5cm of X] (tmp0);
        \vertex[above=0.57cm of tmp0] (x);
        \vertex[left=0.25cm of X] (tmp1);
        \vertex[left=0.75cm of x] (tmp2);
        \vertex[above=1cm of tmp] (e2) {${{ external[2]|texname(generic,True) }} {{momenta[2]}}$};
        \vertex[below=1cm of tmp] (e3) {${{ external[1]|texname(generic,True) }} {{momenta[1]}}$};
        \diagram*{(e1)  -- [{{external[0]|line}}] (X),
                  (e3)  -- [{{external[1]|line}}] (X),
                  (X)  -- [{{internal[0]|line}}, edge label'=${{internal[0]|texname(generic)}}$] (x),
                  (x)  -- [{{internal[1]|line}}, edge label'=${{internal[1]|texname(generic)}}$] (tmp2),
                  (e2)  -- [{{external[2]|line}}] (x)};
        \draw[{{internal[2]|line}}] (tmp2) arc [start angle=0, end angle=360, radius=0.3cm] node[label={[xshift=-1.0cm,yshift=-0.3cm]${{internal[2]|texname(generic)}}$}] {};
    \end{feynman}
\end{tikzpicture}

""")

skeletons[3]['ThreePointCT'] = Template(r"""
\begin{tikzpicture}[baseline=0]
    \begin{feynman}
        \vertex[crossed dot] (X) {};
        \vertex[left=1cm of X] (e1) {${{ external[0]|texname(generic,True) }}$};
        \vertex[right=1cm of X] (tmp);
        \vertex[above=1cm of tmp] (e2) {${{ external[1]|texname(generic,True) }}$};
        \vertex[below=1cm of tmp] (e3) {${{ external[2]|texname(generic,True) }}$};
        \diagram*{(e1)  -- [{{external[0]|line}}] (X),
                  (X)  -- [{{external[1]|line}}] (e2),
                  (X)  -- [{{external[2]|line}}] (e3)};
    \end{feynman}
\end{tikzpicture}

""")

skeletons[3]['ThreePointMCT'] = Template(r"""
\begin{tikzpicture}[baseline=0]
    \begin{feynman}
        \vertex (X);
        \vertex[left=0.5cm of X] (tmp);
        \vertex[crossed dot, above=0.5cm of tmp] (ct) {};
        \vertex[left=1cm of X] (e1) {${{ external[0]|texname(generic,True) }}$};
        \vertex[right=1cm of X] (tmp);
        \vertex[above=1cm of tmp] (e2) {${{ external[1]|texname(generic,True) }}$};
        \vertex[below=1cm of tmp] (e3) {${{ external[2]|texname(generic,True) }}$};
        \diagram*{(e1)  -- [{{external[0]|line}}] (X),
                  (X)  -- [{{external[1]|line}}] (e2),
                  (X)  -- [{{external[2]|line}}] (e3)};
    \end{feynman}
\end{tikzpicture}

""")



import re
import logging

def pythonform(obj):
    """ Returns a string representation of `obj`.
    If `obj` is a list/set/dict, it acts on all elements."""
    if isinstance(obj, str):
        return f"'{obj}'".replace('\\','\\\\')
    if type(obj).__base__.__name__ == "UFOBaseClass":
        return obj.__class__.__name__[0] + "." + repr(obj)
    if isinstance(obj, list):
        return str([pythonform(i) for i in obj]).replace("'","")
    if isinstance(obj, set):
        return str((pythonform(i) for i in obj)).replace("'","")
    if isinstance(obj, dict):
        return str({pythonform(i) : pythonform(j) for i,j in obj.items()}).replace("'","")
    return obj

class UFOError(Exception):
    """Exception raised if when inconsistencies are detected in the UFO model."""
    pass

class UFOBaseClass(object):
    """The class from which all FeynRules classes are derived."""

    require_args = []

    def __init__(self, *args, **options):
        assert len(self.require_args) == len(args)

        for i, name in enumerate(self.require_args):
            setattr(self, name, args[i])

        for (option, value) in options.items():
            setattr(self, option, value)

    def get(self, name):
        return getattr(self, name)

    def set(self, name, value):
        setattr(self, name, value)

    def get_all(self):
        """Return a dictionary containing all the information of the object"""
        return self.__dict__

    def dump(self):
        res = f'\n{repr(self)} = {self.__class__.__name__}(\n'.replace("'","")
        for k,obj in self.__dict__.items():
            res += f'    {k} = {pythonform(obj)},\n'
        res += ')\n'
        return res

    def __str__(self):
        return self.name

    def nice_string(self):
        """ return string with the full information """
        return '\n'.join(['%s \t: %s' % (name, value) for name, value in self.__dict__.items()])

    def __repr__(self):
        replacements = [
            ('+','__plus__'),
            ('-','__minus__'),
            ('@','__at__'),
            ('!','__exclam__'),
            ('?','__quest__'),
            ('*','__star__'),
            ('~','__tilde__')
            ]
        text = self.name
        for orig,sub in replacements:
            text = text.replace(orig,sub)
        return text

all_particles = {}
class Particle(UFOBaseClass):
    """A standard Particle"""

    require_args = [
            'pdg_code', 'name', 'antiname',
            'spin', 'color', 'mass', 'width',
            'texname', 'antitexname', 'charge'
            ]

    require_args_all = [
            'pdg_code', 'name', 'antiname',
            'spin', 'color', 'mass', 'nmass', 'width',
            'texname', 'antitexname','counterterm',
            'charge', 'line', 'propagating',
            'goldstone', 'propagator', 'OS'
            ]

    def __init__(self, pdg_code, name, antiname, spin, color, mass, width, texname,
                 antitexname, charge , line=None, propagating=True, counterterm=None,
                 goldstone=False, propagator=None, OS=False, **options):

        args = (pdg_code, name, antiname, spin, color, mass,
                width, texname,antitexname, float(charge))

        UFOBaseClass.__init__(self, *args,  **options)

        global all_particles
        if name in all_particles:
            logging.warning(f'Particle name convention for {name} not unique!')
        all_particles[name] = self

        self.propagating = propagating
        self.goldstone = goldstone

        self.selfconjugate = (name == antiname)
        if not line:
            self.line = self.find_line_type()
        else:
            self.line = line

        if hasattr(self.mass, 'value'):
            try:
                self.nmass = float(self.mass.value)
            except:
                pass

        if propagator:
            if isinstance(propagator, dict):
                self.propagator = propagator
            else:
                self.propagator = {0: propagator, 1: propagator}

    def find_line_type(self):
        """ find how we draw a line if not defined
        valid output: dashed/straight/wavy/curly/double/swavy/scurly
        """
        spin = self.spin
        color = self.color

        # use default
        if spin == 1:
            return 'dashed'
        elif spin == 2:
            if not self.selfconjugate:
                return 'straight'
            elif color == 1:
                return 'swavy'
            else:
                return 'scurly'
        elif spin == 3:
            if color == 1:
                return 'wavy'

            else:
                return 'curly'
        elif spin == 5:
            return 'double'
        elif spin == -1:
            return 'dotted'
        else:
            return 'dashed' # not supported yet

    def anti(self):
        global all_particles
        if self.selfconjugate:
            return self
            # raise Exception('%s has no anti particle.' % self.name)
        if self.antiname in all_particles:
            return all_particles[self.antiname]
        outdic = {}
        for k,v in self.__dict__.items():
            if k not in self.require_args_all:
                outdic[k] = -v
        if self.color in [1,8]:
            newcolor = self.color
        else:
            newcolor = -self.color

        return Particle(-self.pdg_code, self.antiname, self.name, self.spin, newcolor, self.mass, self.width,
                        self.antitexname, self.texname, -self.charge, self.line, self.propagating, self.goldstone, **outdic)

all_parameters = {}
class Parameter(UFOBaseClass):

    require_args = ['name', 'nature', 'type', 'value', 'texname']
    require_args_all = ['name', 'nature', 'type', 'value', 'nvalue', 'texname']

    def __init__(self, name, nature, type, value, texname, lhablock=None, lhacode=None, **options):

        args = (name,nature,type,value,texname)

        UFOBaseClass.__init__(self, *args)

        args = (name,nature,type,value,texname)

        global all_parameters
        if name in all_parameters:
            logging.warning(f'Parameter name convention for {name} not unique!')
        all_parameters[name] = self

        if (lhablock is None or lhacode is None) and nature == 'external':
            raise Exception('Need LHA information for external parameter "%s".' % name)
        self.lhablock = lhablock
        self.lhacode = lhacode

all_CTparameters = {}
class CTParameter(UFOBaseClass):

    require_args = ['name', 'type', 'value', 'texname']

    def __init__(self, name, type, value, texname):

        args = (name,type,value,texname)

        UFOBaseClass.__init__(self, *args)

        global all_CTparameters
        if name in all_CTparameters:
            logging.warning(f'CT-parameter name convention for {name} not unique!')
        all_CTparameters[name] = self

    def finite(self):
        try:
            return self.value[0]
        except KeyError:
            return 'ZERO'

    def pole(self, x):
        try:
            return self.value[-x]
        except KeyError:
            return 'ZERO'

all_vertices = {}
class Vertex(UFOBaseClass):

    require_args = ['name', 'particles', 'color', 'lorentz', 'couplings']

    def __init__(self, name, particles, color, lorentz, couplings, **opt):
        global all_vertices
        args = (name, particles, color, lorentz, couplings)
        UFOBaseClass.__init__(self, *args, **opt)
        args = (particles,color,lorentz,couplings)
        if name in all_vertices:
            logging.warning(f'Vertex name convention for {name} not unique!')
        all_vertices[name] = self

all_CTvertices = {}
class CTVertex(UFOBaseClass):

    require_args = ['name', 'particles', 'color', 'lorentz', 'couplings', 'type', 'loop_particles']

    def __init__(self, name, particles, color, lorentz, couplings, type, loop_particles, **opt):

        args = (name, particles, color, lorentz, couplings, type, loop_particles)

        UFOBaseClass.__init__(self, *args, **opt)

        args = (particles,color,lorentz,couplings, type, loop_particles)

        global all_CTvertices
        if name in all_vertices:
            logging.warning(f'Vertex-CT name convention for {name} not unique!')
        all_CTvertices[name] = self

all_couplings = {}
class Coupling(UFOBaseClass):

    require_args = ['name', 'value', 'order']

    require_args_all = ['name', 'value', 'nvalue', 'order', 'loop_particles', 'counterterm']

    def __init__(self, name, value, order, **opt):

        args = (name, value, order)
        UFOBaseClass.__init__(self, *args, **opt)
        global all_couplings
        if name in all_couplings:
            logging.warning(f'Coupling name convention for {name} not unique!')
        all_couplings[name] = self

    def value(self):
        return self.pole(0)

    def pole(self, x):
        """ the self.value attribute can be a dictionary directly specifying the Laurent serie using normal
        parameter or just a string which can possibly contain CTparameter defining the Laurent serie."""

        if isinstance(self.value,dict):
            if -x in self.value.keys():
                return self.value[-x]
            else:
                return 'ZERO'

        CTparam = None
        for param in all_CTparameters:
            pattern = re.compile(r"(?P<first>\A|\*|\+|\-|\()(?P<name>"+param.name+r")(?P<second>\Z|\*|\+|\-|\))")
            numberOfMatches = len(pattern.findall(self.value))
            if numberOfMatches == 1:
                if not CTparam:
                    CTparam = param
                else:
                    raise UFOError("UFO does not support yet more than one occurence of CTParameters in the couplings values.")
            elif numberOfMatches > 1:
                raise UFOError("UFO does not support yet more than one occurence of CTParameters in the couplings values.")

        if not CTparam:
            if x == 0:
                return self.value
            else:
                return 'ZERO'
        else:
            if CTparam.pole(x) == 'ZERO':
                return 'ZERO'
            else:
                def substitution(matchedObj):
                    return matchedObj.group('first')+"("+CTparam.pole(x)+")"+matchedObj.group('second')
                pattern = re.compile(r"(?P<first>\A|\*|\+|\-|\()(?P<name>"+CTparam.name+r")(?P<second>\Z|\*|\+|\-|\))")
                return pattern.sub(substitution,self.value)

all_lorentz = {}
class Lorentz(UFOBaseClass):

    require_args = ['name','spins','structure']

    def __init__(self, name, spins, structure='external', **opt):
        args = (name, spins, structure)
        UFOBaseClass.__init__(self, *args, **opt)

        global all_lorentz
        if name in all_lorentz:
            logging.warning(f'Lorentz name convention for {name} not unique!')
        all_lorentz[name] = self

all_orders = {}
class CouplingOrder(object):

    def __init__(self, name, expansion_order, hierarchy, perturbative_expansion = 0):
        global all_orders
        if name in all_orders:
            logging.warning(f'CouplingOrder name convention for {name} not unique!')
        all_orders[name] = self

        self.name = name
        self.expansion_order = expansion_order
        self.hierarchy = hierarchy
        self.perturbative_expansion = perturbative_expansion

all_decays = {}
class Decay(UFOBaseClass):
    require_args = ['particle','partial_widths']

    def __init__(self, particle, partial_widths, **opt):
        args = (particle, partial_widths)
        UFOBaseClass.__init__(self, *args, **opt)

        global all_decays
        if particle in all_decays:
            logging.warning(f'Decay particle convention for {particle} not unique!')
        all_decays[particle] = self

        # Add the information directly to the particle
        particle.partial_widths = partial_widths

all_form_factors = {}
class FormFactor(UFOBaseClass):
    require_args = ['name','type','value']

    def __init__(self, name, type, value, **opt):
        args = (name, type, value)
        UFOBaseClass.__init__(self, *args, **opt)

        global all_form_factors
        if name in all_form_factors:
            logging.warning(f'FormFactor name convention for {name} not unique!')
        all_form_factors[name] = self

all_propagators = {}
class Propagator(UFOBaseClass):

    require_args = ['name','numerator','denominator']

    def __init__(self, name, numerator, denominator=None, **opt):
        args = (name, numerator, denominator)
        UFOBaseClass.__init__(self, *args, **opt)

        global all_propagators
        if name in all_propagators:
            logging.warning(f'Propagator name convention for {name} not unique!')
        all_propagators[name] = self



#!/usr/bin/env python3
import unittest
from anyBSM import anyBSM, loopfunctions
import numpy.testing as test
from cmath import pi, sqrt
import logging
logging.getLogger('anyBSM').setLevel(logging.WARNING)

# different particle restrictions
SM = anyBSM('SM', evaluation='numerical', caching = 0, progress = False, quiet = True)
SM.load_renormalization_scheme('MS')
SM.setparameters()
p = SM.all_particles
higgs = p['h']

v = SM.parameters['vvSM'].nvalue
Mt = SM.all_particles['u3'].nmass
Mh = higgs.nmass
MZ = SM.parameters['MZ'].nvalue
MWp = SM.parameters['MWp'].nvalue
Mb = SM.all_particles['d3'].nmass
Md2 = SM.all_particles['d2'].nmass
Md1 = SM.all_particles['d1'].nmass
Mu2 = SM.all_particles['u2'].nmass
Mu1 = SM.all_particles['u1'].nmass
Me3 = SM.all_particles['e3'].nmass
Me2 = SM.all_particles['e2'].nmass
Me1 = SM.all_particles['e1'].nmass

loopfunctions.set_renscale(Mt)

loopfunctions.MSDR = int(not SM.dimensional_reduction) # legacy

SMall = anyBSM('SM', particles_only=['h','Hp','Ah','Hpc','Wp', 'Z', 'A','gWp', 'gWC', 'gZ','gA','u3'], evaluation='numerical', caching = 0, progress = False,scheme_name="OSFJ", quiet = True)
SMall.setparameters()
SMall.setparameters({"Qren":Mt})
pall = SMall.all_particles

SMscalar = anyBSM('SM', particles_only=['h','Ah','Hp','Hpc'],evaluation='numerical', caching = 0, progress = False,scheme_name="OS", quiet = True)
SMscalar.setparameters()
SMscalar.setparameters({"Qren":Mt})
pscalar = SMscalar.all_particles

SMtop = anyBSM('SM', particles_only=['u3'], evaluation='numerical', caching = 0, progress = False,scheme_name="OS", quiet = True)
SMtop.setparameters()
SMtop.setparameters({"Qren":SMtop.all_particles['u3'].mass})
ptop = SMtop.all_particles

SMtb = anyBSM('SM', particles_only=['u3','d3'], evaluation='numerical', caching = 0, progress = False,scheme_name="OS", quiet = True)
SMtb.setparameters()
SMtb.setparameters({"Qren":SMtop.all_particles['u3'].mass})

SMboson = anyBSM('SM', particles_only=['h','Hp','Ah','Hpc','Wp','Z','gWp','gWC', 'gZ'], evaluation='numerical', caching = 0, progress = False,scheme_name="OS", quiet = True)
SMboson.setparameters()
SMboson.setparameters({"Qren":Mt})
pboson = SMboson.all_particles

SMgauge = anyBSM('SM', particles_only=['Wp','Z','A'], evaluation='numerical', caching = 0, progress = False,scheme_name="OS", quiet = True)
SMgauge.setparameters()
SMgauge.setparameters({"Qren":Mt})
pgauge = SMgauge.all_particles

SMgauge2 = anyBSM('SM', particles_only=['Z','Ah'], evaluation='numerical', caching = 0, progress = False,scheme_name="OS", quiet = True)
SMgauge2.setparameters()
SMgauge2.setparameters({"Qren":Mt})

SMghost = anyBSM('SM', particles_only=['gWp','gWC', 'gZ','gA'], evaluation='numerical', caching = 0, progress = False,scheme_name="MS", quiet = True)
SMghost.setparameters()
SMghost.setparameters({"Qren":Mt})
pghost = SMghost.all_particles

SMG0 = anyBSM('SM', particles_only=['Ah'], evaluation='numerical', caching = 0, progress = False,scheme_name="OS", quiet = True)
SMG0.setparameters()
SMG0.setparameters({"Qren":Mt})
pG0 = SMG0.all_particles

SMlf = anyBSM('SM', particles_only=['nu1','nu2','nu3','d1','d2','d3','u1','u2','e1','e2','e3'], evaluation='numerical', caching = 0, progress = False,scheme_name="OS", quiet = True)
SMlf.setparameters()
SMlf.setparameters({"Qren":Mt})
plf = SMlf.all_particles

SMfermion = anyBSM('SM', particles_only=['nu1','nu2','nu3','d1','d2','d3','u1','u2','u3','e1','e2','e3'], evaluation='numerical', caching = 0, progress = False,scheme_name="OS", quiet = True)
SMfermion.setparameters()
SMfermion.setparameters({"Qren":Mt})


alpha = 1/(SMall._eval('aEWM1'))
EL = sqrt(4*pi*alpha)
CW2 = SMall._eval('MWp/MZ')**2
SW2 = 1-CW2

loopfunctions.set_renscale(Mt)
a0mh = loopfunctions.A0(Mh**2)
a0mt = loopfunctions.A0(Mt**2)
a0mb = loopfunctions.A0(Mb**2)
a0mz = loopfunctions.A0(MZ**2)
a0mw = loopfunctions.A0(MWp**2)
b0mh = loopfunctions.B0(Mh**2,Mh**2,Mh**2)
b0mt = loopfunctions.B0(Mh**2,Mt**2,Mt**2)
b0mw = loopfunctions.B0(Mh**2,MWp**2,MWp**2)
b0mz = loopfunctions.B0(Mh**2,MZ**2,MZ**2)

def hhh1LsFC():
    return - alpha*EL*Mh**4/(64*pi*MWp**3*SW2*sqrt(SW2))*( 27*loopfunctions.B0(Mh**2,Mh**2,Mh**2)+6*loopfunctions.B0(Mh**2,MWp**2,MWp**2)+3*loopfunctions.B0(Mh**2,MZ**2,MZ**2)+2*Mh**2*(27*loopfunctions.C0(Mh**2,Mh**2,Mh**2,Mh**2,Mh**2,Mh**2)+2*loopfunctions.C0(Mh**2,Mh**2,Mh**2,MWp**2,MWp**2,MWp**2)+loopfunctions.C0(Mh**2,Mh**2,Mh**2,MZ**2,MZ**2,MZ**2)))

def hhh1LvFC():
    return alpha*EL*MWp/(8*pi*CW2**3*sqrt(SW2)*SW2)*(3*CW2*loopfunctions.MSDR*(1+2*CW2**2)-12*CW2**3*loopfunctions.B0(Mh**2,MWp**2,MWp**2)-6*CW2*loopfunctions.B0(Mh**2,MZ**2,MZ**2)-16*CW2**3*MWp**2*loopfunctions.C0(Mh**2,Mh**2,Mh**2, MWp**2,MWp**2,MWp**2)-8*MWp**2*loopfunctions.C0(Mh**2,Mh**2,Mh**2,MZ**2,MZ**2,MZ**2))

def hhh1LuFC():
    return alpha*EL/(16*pi*CW2*SW2*sqrt(CW2*SW2))*(2*sqrt(CW2)*CW2*MWp**3*loopfunctions.C0(Mh**2,Mh**2,Mh**2,MWp**2,MWp**2,MWp**2)+MZ**3*loopfunctions.C0(Mh**2,Mh**2,Mh**2,MZ**2,MZ**2,MZ**2))

def hhh1LhFC():
    return - 27*alpha*EL*Mh**4/(64*pi*MWp**3*sqrt(SW2)*SW2)*( loopfunctions.B0(Mh**2,Mh**2,Mh**2)+2*Mh**2*loopfunctions.C0(Mh**2,Mh**2,Mh**2,Mh**2,Mh**2,Mh**2))

def hhh1LGZFC():
    return - alpha*EL/(64*pi*MWp**3*sqrt(SW2)*SW2)*(3*(Mh**4-2*Mh**2*MZ**2+12*MZ**4)*loopfunctions.B0(Mh**2,MZ**2,MZ**2)+2*(Mh**6-2*Mh**4*MZ**2-16*Mh**2*MZ**4+26*MZ**6)*loopfunctions.C0(Mh**2,Mh**2,Mh**2,MZ**2,MZ**2,MZ**2)-6*MZ**2*(4*loopfunctions.MSDR*MZ**2+Mh**2*(Mh**2+2*MZ**2)*(loopfunctions.C1(Mh**2,Mh**2,Mh**2,MZ**2,MZ**2,MZ**2)+loopfunctions.C2(Mh**2,Mh**2,Mh**2,MZ**2,MZ**2,MZ**2))))

def hhh1LGpWFC():
    return alpha*EL/(32*pi*MWp**3*sqrt(SW2)*SW2)*(-3*(Mh**4-2*Mh**2*MWp**2+12*MWp**4)*loopfunctions.B0(Mh**2,MWp**2,MWp**2)-2*(Mh**6-2*Mh**4*MWp**2-16*Mh**2*MWp**4+26*MWp**6)*loopfunctions.C0(Mh**2,Mh**2,Mh**2,MWp**2,MWp**2,MWp**2)+6*MWp**2*(4*loopfunctions.MSDR*MWp**2+Mh**2*(Mh**2+2*MWp**2)*( loopfunctions.C1(Mh**2,Mh**2,Mh**2,MWp**2,MWp**2,MWp**2)+loopfunctions.C2(Mh**2,Mh**2,Mh**2,MWp**2,MWp**2,MWp**2) )))

class numericTest(unittest.TestCase):
    """ Test genuine 1L corrections """

    # Eff-Pot approximation (top quark contrib.)
    def test_top(self):
        # Our result
        OurLambda = SMtop.lambdahhh(momenta=[0,0,0])
        # e.q (1) from hep-ph/0211308 :
        LambdaEffective = OurLambda['treelevel']*(1-Mt**4/(pi*v*Mh)**2)
        test.assert_allclose(LambdaEffective, OurLambda['treelevel'] + OurLambda['genuine'])

    # Full external momentum (top quark contrib.)
    def test_top_full_momentum(self):
        # Our result
        OurLambda = SMtop.lambdahhh(momenta=['auto'])
        # analytical result
        p12 = Mh**2
        p22 = Mh**2
        p32 = Mh**2
        p2p3 = (p12 - p22 - p32)/2 # noqa: F841
        loopfunctions.pyCollier.set_renscale(Mt**2) # noqa: F841
        B0 = loopfunctions.pyCollier.b0 # noqa: F841
        C0 = loopfunctions.pyCollier.c0 # noqa: F841
        C1 = loopfunctions.pyCollier.c1 # noqa: F841
        C2 = loopfunctions.pyCollier.c2 # noqa: F841
        nloshiftanalytic = eval("""2*(1/(16*v**3*pi**2))*6*Mt**4*(
            6*B0(p32, Mt**2, Mt**2)
            + (p12 + p22 - p32 + 8*Mt**2)*C0(p22, p32, p12, Mt**2, Mt**2, Mt**2)
            + 2*(p12 + 3*p22 - p32)     *C1(p22, p32, p12, Mt**2, Mt**2, Mt**2)
            + 2*(3*p12 + p22 - p32)     *C2(p22, p32, p12, Mt**2, Mt**2, Mt**2)
        )""")
        test.assert_allclose(nloshiftanalytic, OurLambda['genuine'])

    # Test scalar contributions
    def test_scalar(self):
        hhh1LsanyBSM = SMscalar.lambdahhh(simplify = False, momenta = ['auto'])['genuine']
        test.assert_allclose(hhh1LsanyBSM, hhh1LsFC())

    # Test gauge contributions
    def test_gauge(self):
        hhh1LvanyBSM = SMgauge.lambdahhh(simplify = False, momenta = ['auto'])['genuine']
        test.assert_allclose(hhh1LvanyBSM, hhh1LvFC())

    # Test scalar+gauge contributions
    def test_scalar_gauge(self):
        hhh1LsvanyBSM = SMboson.lambdahhh(simplify = False, momenta = ['auto'])['genuine']
        test.assert_allclose(hhh1LsvanyBSM, hhh1LhFC()+hhh1LGZFC()+hhh1LGpWFC()+hhh1LuFC())

    # Test Goldstone + Z contributions
    def test_GZ(self):
        hhh1LGZanyBSM = SMgauge2.lambdahhh(simplify = False, momenta = ['auto'])['genuine']
        test.assert_allclose(hhh1LGZanyBSM, hhh1LGZFC())

    # Test Ghost contributions
    def test_ghost(self):
        hhh1LuanyBSM = SMghost.lambdahhh(simplify = False, momenta = ['auto'])['genuine']
        test.assert_allclose(hhh1LuanyBSM, hhh1LuFC())

    # Test Goldstone only contributions
    def test_goldstone(self):
        hhh1LGanyBSM = SMG0.lambdahhh(simplify = False, momenta = ['auto'])['genuine']
        hhhGGGFC = alpha*EL*Mh**6*MZ**2/(32*pi*MWp**3*sqrt(SW2)*(MWp**2-MZ**2))*loopfunctions.C0(Mh**2,Mh**2,Mh**2,MZ**2,MZ**2,MZ**2)
        hhhGGFC = - 3*alpha*EL*Mh**4/(64*pi*MWp**3*sqrt(SW2)*SW2)*loopfunctions.B0(Mh**2,MZ**2,MZ**2)
        test.assert_allclose(hhh1LGanyBSM, hhhGGGFC + hhhGGFC)

    # Test tadpole contributions
    def test_tadpole(self):
        hhhTAanyBSM = SMall.lambdahhh(simplify = False, momenta = ['auto'])['tads']
        hhhTAFC = 3*alpha*EL/(64*pi*MWp**3*SW2*sqrt(SW2))*(Mh**2*(3*a0mh+2*a0mw+a0mz)-24*Mt**2*a0mt + 12*MWp**2*a0mw+6*MZ**2*a0mz-4*loopfunctions.MSDR*(2*MWp**4+MZ**4))
        test.assert_allclose(hhhTAanyBSM, hhhTAFC)

    # Test WFR contributions
    # internal cross-check
    def test_wfr_internal(self):
        dZH = sum(SMall.process('h','h',derivative = True, momenta=[Mh**2], simplify = False).values())
        hhhWFRdZ = 3/2*3*Mh**2/v*dZH
        hhhWFRanyBSM = SMall.lambdahhh(simplify = False, momenta = ['auto'])['wfr']
        test.assert_allclose(hhhWFRdZ, hhhWFRanyBSM)

    # external cross-check
    def test_wfr_external(self):
        hhhWFRanyBSM = SMall.lambdahhh(simplify = False, momenta = ['auto'])['wfr']
        hhhWFRsfFC = - 3/2*3*Mh**2/v*alpha/(32*pi*MWp**2*SW2)*(Mh**4*(9*loopfunctions.dB0(Mh**2,Mh**2,Mh**2)+2*loopfunctions.dB0(Mh**2,MWp**2,MWp**2)+loopfunctions.dB0(Mh**2,MZ**2, MZ**2))-24*Mt**2*(-loopfunctions.B0(Mh**2,Mt**2,Mt**2)/2 + 2*Mt**2*loopfunctions.dB0(Mh**2,Mt**2,Mt**2)-Mh**2/2*loopfunctions.dB0(Mh**2,Mt**2,Mt**2)))
        hhhWFRgbFC = 3/2*3*Mh**2/v*alpha/(16*pi*MWp**2*SW2)*(-16*MWp**4*loopfunctions.dB0(Mh**2,MWp**2,MWp**2)-8*MZ**4*loopfunctions.dB0(Mh**2,MZ**2,MZ**2))
        hhhWFRghostFC = 3/2*3*Mh**2/v*alpha/(16*pi*MWp**2*SW2)*(2*MWp**4*loopfunctions.dB0(Mh**2,MWp**2,MWp**2)+MZ**4*loopfunctions.dB0(Mh**2,MZ**2,MZ**2))
        hhhWFRsgbFC = 3/2*3*Mh**2/v*alpha/(16*pi*MWp**2*SW2)*(4*MWp**2*loopfunctions.B0(Mh**2,MWp**2, MWp**2)+2*MZ**2*loopfunctions.B0(Mh**2,MZ**2,MZ**2)+4*Mh**2*MWp**2*loopfunctions.dB0(Mh**2,MWp**2,MWp**2)+2*MWp**4*loopfunctions.dB0(Mh**2,MWp**2,MWp**2)+2*Mh**2*MZ**2*loopfunctions.dB0(Mh**2,MZ**2,MZ**2)+MZ**4*loopfunctions.dB0(Mh**2,MZ**2,MZ**2))
        test.assert_allclose(hhhWFRanyBSM, hhhWFRsfFC + hhhWFRgbFC + hhhWFRghostFC + hhhWFRsgbFC)

    def test_Higgs_mass_ren(self):
        """ test Higgs mass renormalization """
        hhhWFRanyBSM = SMall.lambdahhh(simplify = False, momenta = ['auto'])['massren']
        hhhmassrenbyhand = - 3/v*sum(SMall.process('h','h', momenta=[Mh**2], simplify = False).values())
        test.assert_allclose(hhhWFRanyBSM, hhhmassrenbyhand)

    def test_se_hh_A(self):
        """ test Higgs self-energy """
        PihhAtotanyBSM = sum(SMall.process('h','h',momenta=[Mh**2],only_topologies=['TwoPointA']).values())
        PihhAsFC = - alpha/(32*pi*SW2*MWp**2)*Mh**2*(3*a0mh+2*a0mw+a0mz)
        PihhAvFC = alpha/(8*pi*CW2*SW2)*(loopfunctions.MSDR*(2*CW2*MWp**2+MZ**2)-4*CW2*a0mw - 2*a0mz)
        test.assert_allclose(PihhAtotanyBSM, PihhAsFC + PihhAvFC)

    def test_se_hh_B_scalar(self):
        PihhBsanyBSM = sum(SMscalar.process('h','h',momenta=[Mh**2] , only_topologies=['TwoPointB']).values())
        PihhBsFC  = - alpha*Mh**4/(32*pi*MWp**2*SW2)*(9*b0mh+2*b0mw+b0mz)
        test.assert_allclose(PihhBsanyBSM, PihhBsFC)

    def test_se_hh_B_top(self):
        PihhBtanyBSM = sum(SMtop.process('h','h',momenta=[Mh**2], only_topologies=['TwoPointB']).values())
        PihhBtFC = 12*alpha*Mt**2/(16*pi*MWp**2*SW2)*(a0mt+(2*Mt**2-Mh**2/2)*b0mt)
        test.assert_allclose(PihhBtanyBSM, PihhBtFC)

    def test_se_hh_B_pure_gauge(self):
        PihhBvanyBSM = sum(SMgauge.process('h','h',momenta=[Mh**2], only_topologies=['TwoPointB']).values())
        PihhBvFC = alpha*MWp**2/(4*pi*CW2**2*SW2)*(loopfunctions.MSDR+2*CW2**2*loopfunctions.MSDR - 4*CW2**2*b0mw-2*b0mz)
        test.assert_allclose(PihhBvanyBSM, PihhBvFC)

    def test_se_hh_B_ghost(self):
        PihhBuanyBSM = sum(SMghost.process('h','h',momenta=[Mh**2], only_topologies=['TwoPointB']).values())
        PihhBuFC = alpha/(16*pi*CW2*SW2)*(2*CW2*MWp**2*b0mw+MZ**2*b0mz)
        test.assert_allclose(PihhBuanyBSM, PihhBuFC)

    def test_se_hh_B_total(self):
        PihhBtotanyBSM = sum(SMall.process('h','h',momenta=[Mh**2], only_topologies=['TwoPointB']).values())
        PihhBtFC = 12*alpha*Mt**2/(16*pi*MWp**2*SW2)*(a0mt+(2*Mt**2-Mh**2/2)*b0mt)
        PihhBsFC  = - alpha*Mh**4/(32*pi*MWp**2*SW2)*(9*b0mh+2*b0mw+b0mz)
        PihhBgaugeFC = alpha/(16*pi*CW2**2*SW2)*(4*loopfunctions.MSDR*MWp**2 + 8*CW2**2*loopfunctions.MSDR*MWp**2+2*CW2**2*a0mw+CW2*a0mz+2*CW2**2*Mh**2*b0mw-14*CW2**2*MWp**2*b0mw+CW2*Mh**2*b0mz-8*MWp**2*b0mz+CW2*MZ**2*b0mz + 2*CW2**2*Mh**2*b0mw+CW2*Mh**2*b0mz)
        PihhBuFC = alpha/(16*pi*CW2*SW2)*(2*CW2*MWp**2*b0mw+MZ**2*b0mz)
        test.assert_allclose(PihhBtotanyBSM, PihhBsFC+PihhBtFC+PihhBgaugeFC+PihhBuFC)

    def test_se_hh_tad(self):
        PihhTAtotanyBSM = sum(SMall.process('h','h',momenta=[Mh**2], only_topologies=['TwoPointTA']).values())
        PihhTAsFC  = 3*alpha*Mh**2*(3*a0mh+2*a0mw+a0mz)/(32*pi*MWp**2*SW2)
        PihhTAtFC  = - 36*alpha*Mt**2*a0mt/(16*pi*MWp**2*SW2)
        PihhTAvFC  = - 6*alpha/(16*pi*CW2*SW2)*(loopfunctions.MSDR*(2*CW2*MWp**2+MZ**2)-4*CW2*a0mw-2*a0mz)
        PihhTAuFC  = - 3*alpha/(16*pi*SW2*MWp**2)*(2*MWp**2*a0mw+MZ**2*a0mz)
        test.assert_allclose(PihhTAtotanyBSM, PihhTAsFC+PihhTAtFC+PihhTAvFC+PihhTAuFC)

    def test_se_GG_tot(self):
        """ test Goldstone self-energy """
        PiGGAtotanyBSM = sum(SM.process('Ah','Ah',momenta=[MZ**2], only_topologies=['TwoPointA']).values())
        PiGGAtotFC = alpha/(32*pi*MWp**2*SW2)*(8*loopfunctions.MSDR*MWp**4 + 4*loopfunctions.MSDR*MZ**4-Mh**2*a0mh-2*(Mh**2+8*MWp**2)*a0mw - (3*Mh**2+8*MZ**2)*a0mz)
        test.assert_allclose(PiGGAtotanyBSM, PiGGAtotFC)

    def test_se_GG_scalar(self):
        PiGGBsanyBSM = sum(SMscalar.process('Ah', 'Ah',momenta=[MZ**2] ,only_topologies=['TwoPointB']).values())
        PiGGBsFC = - alpha*Mh**4/(16*pi*MWp**2*SW2)*loopfunctions.B0(MZ**2,Mh**2, MZ**2)
        test.assert_allclose(PiGGBsanyBSM, PiGGBsFC)

    def test_se_GG_scalar_gauge(self):
        PiGGBsvanyBSM = sum(SMboson.process('Ah', 'Ah',momenta=[MZ**2] ,only_topologies=['TwoPointB']).values())
        PiGGBsanyBSM = sum(SMscalar.process('Ah', 'Ah',momenta=[MZ**2] ,only_topologies=['TwoPointB']).values())
        PiGGBuanyBSM = sum(SMghost.process('Ah', 'Ah',momenta=[MZ**2] ,only_topologies=['TwoPointB']).values())
        PiGGBsvFC = alpha/(16*pi*MWp**2*SW2)*(-MZ**2*a0mh+2*MWp**2*a0mw + 2*MZ**2*a0mz + 2*Mh**2*MZ**2*loopfunctions.B0(MZ**2,Mh**2,MZ**2)+MZ**4*loopfunctions.B0(MZ**2,Mh**2,MZ**2)+2*MWp**4*loopfunctions.B0(MZ**2,MWp**2,MWp**2)+4*MWp**2*MZ**2*loopfunctions.B0(MZ**2,MWp**2,MWp**2))
        test.assert_allclose(PiGGBsvanyBSM-PiGGBsanyBSM-PiGGBuanyBSM, PiGGBsvFC)

    def test_se_GG_top(self):
        PiGGBtopanyBSM = sum(SMtop.process('Ah', 'Ah',momenta=[MZ**2] ,only_topologies=['TwoPointB']).values())
        PiGGBtopFC = 12*alpha*Mt**2/(16*pi*MWp**2*SW2)*(a0mt-MZ**2/2*loopfunctions.B0(MZ**2,Mt**2,Mt**2))
        test.assert_allclose(PiGGBtopanyBSM, PiGGBtopFC)

    def test_se_GG_ghost(self):
        PiGGBuanyBSM = sum(SMghost.process('Ah', 'Ah',momenta=[MZ**2] ,only_topologies=['TwoPointB']).values())
        PiGGBuFC = - alpha*MWp**2/(8*pi*SW2)*loopfunctions.B0(MZ**2,MWp**2,MWp**2)
        test.assert_allclose(PiGGBuanyBSM, PiGGBuFC)

    def test_se_GG_tad(self):
        PiGGTAtotanyBSM = sum(SMall.process('Ah','Ah',momenta=[MZ**2], only_topologies=['TwoPointTA']).values())
        PiGGTAtotFC = - alpha/(32*pi*MWp**2*SW2)*(8*loopfunctions.MSDR*MWp**4 + 4*loopfunctions.MSDR*MZ**4-3*Mh**2*a0mh + 24*Mt**2*a0mt-2*(Mh**2+6*MWp**2)*a0mw - (Mh**2+6*MZ**2)*a0mz)
        test.assert_allclose(PiGGTAtotanyBSM, PiGGTAtotFC)

    def test_se_ZZ_A_scalar(self):
        """ test Z self-energy """
        PiZZAsanyBSM = sum(SMscalar.process('Z','Z',momenta=[MZ**2], only_topologies=['TwoPointA']).values())
        PiZZAsFC = alpha/(16*pi*CW2*SW2)*(a0mh+2*(CW2-SW2)**2*a0mw+a0mz)
        test.assert_allclose(PiZZAsanyBSM, PiZZAsFC)

    def test_se_ZZ_A_vector(self):
        PiZZAvanyBSM = sum(SMgauge.process('Z','Z',momenta=[MZ**2], only_topologies=['TwoPointA']).values())
        PiZZAvFC = - alpha*CW2/(2*pi*SW2)*(2*loopfunctions.MSDR*MWp**2-3*a0mw)
        test.assert_allclose(PiZZAvanyBSM, PiZZAvFC)

    def test_se_ZZ_B_scalar(self):
        PiZZBsanyBSM = sum(SMscalar.process('Z','Z',momenta=[MZ**2], only_topologies=['TwoPointB']).values())
        PiZZBsFC = - alpha/(4*pi*CW2*SW2)*(loopfunctions.B00(MZ**2,Mh**2,MZ**2)+(CW2-SW2)**2*loopfunctions.B00(MZ**2,MWp**2,MWp**2))
        test.assert_allclose(PiZZBsanyBSM, PiZZBsFC)

    def test_se_ZZ_B_top(self):
        PiZZBtanyBSM = sum(SMtop.process('Z','Z',momenta=[MZ**2], only_topologies=['TwoPointB']).values())
        PiZZBtFC = - alpha/(24*pi*CW2*SW2)*((9-24*SW2+32*SW2**2)*a0mt + 9*Mt**2*loopfunctions.B0(MZ**2,Mt**2,Mt**2)-(9-24*SW2+32*SW2**2)*(2*loopfunctions.B00(MZ**2,Mt**2,Mt**2)+MZ**2/2*loopfunctions.B0(MZ**2,Mt**2,Mt**2)))
        test.assert_allclose(PiZZBtanyBSM, PiZZBtFC)

    def test_se_ZZ_B_lf(self):
        PiZZBlfanyBSM = sum(SMlf.process('Z','Z',momenta=[MZ**2], only_topologies=['TwoPointB']).values())
        PiZZBbFC = - alpha/(24*pi*CW2*SW2)*((9-12*SW2+8*SW2**2)*loopfunctions.A0(Mb**2)+9*Mb**2*loopfunctions.B0(MZ**2,Mb**2,Mb**2)-2*(9-12*SW2+8*SW2**2)*loopfunctions.B00(MZ**2,Mb**2,Mb**2)+MZ**2*(-1/2*(9-12*SW2+8*SW2**2)*loopfunctions.B0(MZ**2,Mb**2,Mb**2)))
        PiZZBotherlfFC = - alpha/(24*pi*CW2*SW2)*((9-12*SW2+8*SW2**2)*loopfunctions.A0(Md2**2)+9*Md2**2*loopfunctions.B0(MZ**2,Md2**2,Md2**2)-2*(9-12*SW2+8*SW2**2)*loopfunctions.B00(MZ**2,Md2**2,Md2**2)+MZ**2*(-1/2*(9-12*SW2+8*SW2**2)*loopfunctions.B0(MZ**2,Md2**2,Md2**2)))-alpha/(24*pi*CW2*SW2)*((9-12*SW2+8*SW2**2)*loopfunctions.A0(Md1**2)+9*Md1**2*loopfunctions.B0(MZ**2,Md1**2,Md1**2)-2*(9-12*SW2+8*SW2**2)*loopfunctions.B00(MZ**2,Md1**2,Md1**2)+MZ**2*(-1/2*(9-12*SW2+8*SW2**2)*loopfunctions.B0(MZ**2,Md1**2,Md1**2)))-alpha/(24*pi*CW2*SW2)*((9-24*SW2+32*SW2**2)*loopfunctions.A0(Mu2**2)+9*Mu2**2*loopfunctions.B0(MZ**2,Mu2**2,Mu2**2)-(9-24*SW2+32*SW2**2)*(2*loopfunctions.B00(MZ**2,Mu2**2,Mu2**2)+MZ**2/2*loopfunctions.B0(MZ**2,Mu2**2,Mu2**2)))-alpha/(24*pi*CW2*SW2)*((9-24*SW2+32*SW2**2)*loopfunctions.A0(Mu1**2)+9*Mu1**2*loopfunctions.B0(MZ**2,Mu1**2,Mu1**2)-(9-24*SW2+32*SW2**2)*(2*loopfunctions.B00(MZ**2,Mu1**2,Mu1**2)+MZ**2/2*loopfunctions.B0(MZ**2,Mu1**2,Mu1**2)))+6*alpha/(16*pi*CW2*SW2)*(2*loopfunctions.B00(MZ**2,0,0)+MZ**2/2*loopfunctions.B0(MZ**2,0,0))-alpha/(8*pi*CW2*SW2)*((1-4*SW2+8*SW2**2)*loopfunctions.A0(Me3**2)+Me3**2*loopfunctions.B0(MZ**2,Me3**2,Me3**2)-(1-4*SW2+8*SW2**2)*(2*loopfunctions.B00(MZ**2,Me3**2,Me3**2)+MZ**2/2*loopfunctions.B0(MZ**2,Me3**2,Me3**2)))-alpha/(8*pi*CW2*SW2)*((1-4*SW2+8*SW2**2)*loopfunctions.A0(Me2**2)+Me2**2*loopfunctions.B0(MZ**2,Me2**2,Me2**2)-(1-4*SW2+8*SW2**2)*(2*loopfunctions.B00(MZ**2,Me2**2,Me2**2)+MZ**2/2*loopfunctions.B0(MZ**2,Me2**2,Me2**2)))-alpha/(8*pi*CW2*SW2)*((1-4*SW2+8*SW2**2)*loopfunctions.A0(Me1**2)+Me1**2*loopfunctions.B0(MZ**2,Me1**2,Me1**2)-(1-4*SW2+8*SW2**2)*(2*loopfunctions.B00(MZ**2,Me1**2,Me1**2)+MZ**2/2*loopfunctions.B0(MZ**2,Me1**2,Me1**2)))
        test.assert_allclose(PiZZBlfanyBSM, PiZZBbFC+PiZZBotherlfFC)

    def test_se_ZZ_B_gauge(self):
        PiZZBvanyBSM = sum(SMgauge.process('Z','Z',momenta=[MZ**2], only_topologies=['TwoPointB']).values())
        PiZZBvFC = - alpha*CW2/(12*pi*SW2)*(loopfunctions.MSDR*(-12*MWp**2+2*MZ**2)+6*a0mw + 3*(2*MWp**2+5*MZ**2)*loopfunctions.B0(MZ**2,MWp**2,MWp**2)+30*loopfunctions.B00(MZ**2,MWp**2,MWp**2)-3*MZ**2*loopfunctions.B0(MZ**2,MWp**2,MWp**2))
        test.assert_allclose(PiZZBvanyBSM, PiZZBvFC)

    def test_se_ZZ_B_ghost(self):
        PiZZBuanyBSM = sum(SMghost.process('Z','Z',momenta=[MZ**2], only_topologies=['TwoPointB']).values())
        PiZZBuFC = alpha*CW2/(2*pi*SW2)*loopfunctions.B00(MZ**2,MWp**2,MWp**2)
        test.assert_allclose(PiZZBuanyBSM, PiZZBuFC)

    def test_se_ZZ_B_total(self):
        PiZZBtotanyBSM = sum(SM.process('Z','Z',momenta=[MZ**2], only_topologies=['TwoPointB']).values())
        PiZZBsFC = - alpha/(4*pi*CW2*SW2)*(loopfunctions.B00(MZ**2,Mh**2,MZ**2)+(CW2-SW2)**2*loopfunctions.B00(MZ**2,MWp**2,MWp**2))
        PiZZBtFC = - alpha/(24*pi*CW2*SW2)*((9-24*SW2+32*SW2**2)*a0mt + 9*Mt**2*loopfunctions.B0(MZ**2,Mt**2,Mt**2)-(9-24*SW2+32*SW2**2)*(2*loopfunctions.B00(MZ**2,Mt**2,Mt**2)+MZ**2/2*loopfunctions.B0(MZ**2,Mt**2,Mt**2)))
        PiZZBvFC = - alpha*CW2/(12*pi*SW2)*(loopfunctions.MSDR*(-12*MWp**2+2*MZ**2)+6*a0mw + 3*(2*MWp**2+5*MZ**2)*loopfunctions.B0(MZ**2,MWp**2,MWp**2)+30*loopfunctions.B00(MZ**2,MWp**2,MWp**2)-3*MZ**2*loopfunctions.B0(MZ**2,MWp**2,MWp**2))
        PiZZBsvFC = alpha*MWp**2/(4*pi*CW2**2*SW2)*(loopfunctions.B0(MZ**2,Mh**2,MZ**2)+2*CW2*SW2**2*loopfunctions.B0(MZ**2,MWp**2,MWp**2))
        PiZZBuFC = alpha*CW2/(2*pi*SW2)*loopfunctions.B00(MZ**2,MWp**2,MWp**2)
        PiZZBbFC = - alpha/(24*pi*CW2*SW2)*((9-12*SW2+8*SW2**2)*loopfunctions.A0(Mb**2)+9*Mb**2*loopfunctions.B0(MZ**2,Mb**2,Mb**2)-2*(9-12*SW2+8*SW2**2)*loopfunctions.B00(MZ**2,Mb**2,Mb**2)+MZ**2*(-1/2*(9-12*SW2+8*SW2**2)*loopfunctions.B0(MZ**2,Mb**2,Mb**2)))
        PiZZBotherlfFC = - alpha/(24*pi*CW2*SW2)*((9-12*SW2+8*SW2**2)*loopfunctions.A0(Md2**2)+9*Md2**2*loopfunctions.B0(MZ**2,Md2**2,Md2**2)-2*(9-12*SW2+8*SW2**2)*loopfunctions.B00(MZ**2,Md2**2,Md2**2)+MZ**2*(-1/2*(9-12*SW2+8*SW2**2)*loopfunctions.B0(MZ**2,Md2**2,Md2**2)))-alpha/(24*pi*CW2*SW2)*((9-12*SW2+8*SW2**2)*loopfunctions.A0(Md1**2)+9*Md1**2*loopfunctions.B0(MZ**2,Md1**2,Md1**2)-2*(9-12*SW2+8*SW2**2)*loopfunctions.B00(MZ**2,Md1**2,Md1**2)+MZ**2*(-1/2*(9-12*SW2+8*SW2**2)*loopfunctions.B0(MZ**2,Md1**2,Md1**2)))-alpha/(24*pi*CW2*SW2)*((9-24*SW2+32*SW2**2)*loopfunctions.A0(Mu2**2)+9*Mu2**2*loopfunctions.B0(MZ**2,Mu2**2,Mu2**2)-(9-24*SW2+32*SW2**2)*(2*loopfunctions.B00(MZ**2,Mu2**2,Mu2**2)+MZ**2/2*loopfunctions.B0(MZ**2,Mu2**2,Mu2**2)))-alpha/(24*pi*CW2*SW2)*((9-24*SW2+32*SW2**2)*loopfunctions.A0(Mu1**2)+9*Mu1**2*loopfunctions.B0(MZ**2,Mu1**2,Mu1**2)-(9-24*SW2+32*SW2**2)*(2*loopfunctions.B00(MZ**2,Mu1**2,Mu1**2)+MZ**2/2*loopfunctions.B0(MZ**2,Mu1**2,Mu1**2)))+6*alpha/(16*pi*CW2*SW2)*(2*loopfunctions.B00(MZ**2,0,0)+MZ**2/2*loopfunctions.B0(MZ**2,0,0))-alpha/(8*pi*CW2*SW2)*((1-4*SW2+8*SW2**2)*loopfunctions.A0(Me3**2)+Me3**2*loopfunctions.B0(MZ**2,Me3**2,Me3**2)-(1-4*SW2+8*SW2**2)*(2*loopfunctions.B00(MZ**2,Me3**2,Me3**2)+MZ**2/2*loopfunctions.B0(MZ**2,Me3**2,Me3**2)))-alpha/(8*pi*CW2*SW2)*((1-4*SW2+8*SW2**2)*loopfunctions.A0(Me2**2)+Me2**2*loopfunctions.B0(MZ**2,Me2**2,Me2**2)-(1-4*SW2+8*SW2**2)*(2*loopfunctions.B00(MZ**2,Me2**2,Me2**2)+MZ**2/2*loopfunctions.B0(MZ**2,Me2**2,Me2**2)))-alpha/(8*pi*CW2*SW2)*((1-4*SW2+8*SW2**2)*loopfunctions.A0(Me1**2)+Me1**2*loopfunctions.B0(MZ**2,Me1**2,Me1**2)-(1-4*SW2+8*SW2**2)*(2*loopfunctions.B00(MZ**2,Me1**2,Me1**2)+MZ**2/2*loopfunctions.B0(MZ**2,Me1**2,Me1**2)))
        test.assert_allclose(PiZZBtotanyBSM, PiZZBvFC+PiZZBsFC+PiZZBsvFC+PiZZBuFC + PiZZBtFC+PiZZBbFC+PiZZBotherlfFC)

    def test_se_ZZ_tadpole(self):
        PiZZTAtotanyBSM = sum(SMall.process('Z','Z',momenta=[MZ**2], only_topologies=['TwoPointTA']).values())
        PiZZTAtotFC = - alpha/(16*pi*sqrt(CW2)*CW2**2*Mh**2*SW2)*( - 8*sqrt(CW2)*CW2*loopfunctions.MSDR*MWp**4 - 4*sqrt(CW2)*MWp**2*MZ**2+3*sqrt(CW2)*CW2*Mh**2*a0mh - 24*sqrt(CW2)*CW2*Mt**2*a0mt+2*sqrt(CW2)*CW2*Mh**2*a0mw + 12*sqrt(CW2)*CW2*MWp**2*a0mw+sqrt(CW2)*CW2*Mh**2*a0mz + 8*sqrt(CW2)*MWp**2*a0mz-2*CW2*MWp*MZ*a0mz)
        test.assert_allclose(PiZZTAtotanyBSM, PiZZTAtotFC)

    def test_se_WW_A_scalar(self):
        """ test W self-energy """
        PiWWAsanyBSM = sum(SMscalar.process('Wp', 'Wpc',momenta=[MWp**2],only_topologies=["TwoPointA"]).values())
        PiWWAsFC = alpha/(16*pi*SW2)*(a0mh+2*a0mw+a0mz)
        test.assert_allclose(PiWWAsanyBSM, PiWWAsFC)

    def test_se_WW_A_gauge(self):
        PiWWAvanyBSM = sum(SMgauge.process('Wp', 'Wpc',momenta=[MWp**2],only_topologies=["TwoPointA"]).values())
        PiWWAWFC = - alpha/(4*pi*SW2)*(2*loopfunctions.MSDR*(MWp**2)-3*a0mw)
        PiWWAZFC = - alpha/(4*pi*SW2)*(2*loopfunctions.MSDR*(CW2*MZ**2)-3*CW2*a0mz)
        test.assert_allclose(PiWWAvanyBSM, PiWWAZFC+PiWWAWFC)

    def test_se_WW_B_scalar(self):
        PiWWBsanyBSM = sum(SMscalar.process('Wp', 'Wpc',momenta=[MWp**2],only_topologies=["TwoPointB"]).values())
        PiWWBsFC = - alpha/(4*pi*SW2)*(loopfunctions.B00(MWp**2,Mh**2,MWp**2)+loopfunctions.B00(MWp**2,MWp**2,MZ**2))
        test.assert_allclose(PiWWBsanyBSM, PiWWBsFC)

    def test_se_WW_B_topbot(self):
        PiWWBtbanyBSM = sum(SMtb.process('Wp', 'Wpc',momenta=[MWp**2],only_topologies=["TwoPointB"]).values())
        PiWWBtbFC = - 6*alpha/(16*pi*SW2)*(loopfunctions.A0(Mb**2)+a0mt+(Mb**2+Mt**2-MWp**2)*loopfunctions.B0(MWp**2,Mb**2,Mt**2)-4*loopfunctions.B00(MWp**2,Mb**2,Mt**2))
        test.assert_allclose(PiWWBtbanyBSM, PiWWBtbFC)

    def test_se_WW_B_gauge(self):
        PiWWBgbanyBSM = sum(SMgauge.process('Wp', 'Wpc',momenta=[MWp**2],only_topologies=["TwoPointB"]).values())
        PiWWBgbFC = alpha/(12*pi*SW2)*(4*loopfunctions.MSDR*MWp**2+6*CW2*loopfunctions.MSDR*MZ**2-3*a0mw-3*CW2*a0mz-15*SW2*MWp**2*loopfunctions.B0(MWp**2,0,MWp**2)-15*CW2*MWp**2*loopfunctions.B0(MWp**2,MWp**2,MZ**2)-3*CW2*MZ**2*loopfunctions.B0(MWp**2,MWp**2,MZ**2)-30*SW2*loopfunctions.B00(MWp**2,0,MWp**2)-30*CW2*loopfunctions.B00(MWp**2,MWp**2,MZ**2))
        test.assert_allclose(PiWWBgbanyBSM, PiWWBgbFC)

    def test_se_WW_B_ghost(self):
        PiWWBuanyBSM = sum(SMghost.process('Wp', 'Wpc',momenta=[MWp**2],only_topologies=["TwoPointB"]).values())
        PiWWBuFC = alpha/(2*pi)*(loopfunctions.B00(MWp**2,0,MWp**2)+CW2/SW2*loopfunctions.B00(MWp**2,MWp**2,MZ**2))
        test.assert_allclose(PiWWBuanyBSM, PiWWBuFC)

    def test_se_WW_tad(self):
        PiWWTAtotanyBSM = sum(SMall.process('Wp', 'Wpc',momenta=[MWp**2],only_topologies=["TwoPointTA"]).values())
        PiWWTAtotFC = - alpha/(16*pi*sqrt(CW2)*CW2*SW2*Mh**2)*(-8*sqrt(CW2)*CW2*loopfunctions.MSDR*MWp**4-4*sqrt(CW2)*loopfunctions.MSDR*MWp**2*MZ**2+3*sqrt(CW2)*CW2*Mh**2*a0mh-24*sqrt(CW2)*CW2*Mt**2*a0mt+2*sqrt(CW2)*CW2*Mh**2*a0mw+12*sqrt(CW2)*CW2*MWp**2*a0mw+sqrt(CW2)*CW2*Mh**2*a0mz+8*sqrt(CW2)*MWp**2*a0mz-2*CW2*MWp*MZ*a0mz)
        test.assert_allclose(PiWWTAtotanyBSM, PiWWTAtotFC)

    def test_se_AA_A_scalar(self):
        """ test photon self-energy """
        PiAAAsanyBSM = sum(SMscalar.process('A','A',momenta=[0], only_topologies=["TwoPointA"]).values())
        PiAAAsFC = alpha/(2*pi)*a0mw
        test.assert_allclose(PiAAAsanyBSM, PiAAAsFC)

    def test_se_AA_A_gauge(self):
        PiAAAvanyBSM = sum(SMgauge.process('A','A',momenta=[0], only_topologies=["TwoPointA"]).values())
        PiAAAvFC = - alpha/(2*pi)*(2*loopfunctions.MSDR*MWp**2-3*a0mw)
        test.assert_allclose(PiAAAvanyBSM, PiAAAvFC)

    def test_se_AA_B_gauge(self):
        PiAABbosonanyBSM = sum(SMboson.process('A','A',momenta=[0], only_topologies=["TwoPointB"]).values())
        PiAABuFC = alpha/(2*pi)*loopfunctions.B00(0,MWp**2,MWp**2)
        PiAABsFC = - alpha/(pi)*loopfunctions.B00(0,MWp**2,MWp**2)
        PiAABvFC = alpha/(2*pi)*(2*loopfunctions.MSDR*MWp**2-a0mw-MWp**2*loopfunctions.B0(0,MWp**2,MWp**2)-5*loopfunctions.B00(0,MWp**2,MWp**2))
        PiAABsvFC = alpha/(2*pi)*MWp**2*loopfunctions.B0(0,MWp**2,MWp**2)
        test.assert_allclose(PiAABbosonanyBSM, PiAABsFC+PiAABsvFC+PiAABvFC+PiAABuFC)

    def test_se_AA_B_topbot(self):
        PiAABfanyBSM = sum(SMtb.process('A','A',momenta=[0], only_topologies=["TwoPointB"]).values())
        PiAABfFC = - alpha/(3*pi)*(a0mb+4*a0mt-2*(loopfunctions.B00(0,Mb**2,Mb**2)+4*loopfunctions.B00(0,Mt**2,Mt**2)))
        test.assert_allclose(PiAABfanyBSM, PiAABfFC, rtol = 0, atol = 1e-10)

    def test_se_AA_B_ghost(self):
        PiAABuanyBSM = sum(SMghost.process('A','A',momenta=[0], only_topologies=["TwoPointB"]).values())
        PiAABuFC = alpha/(2*pi)*loopfunctions.B00(0,MWp**2,MWp**2)
        test.assert_allclose(PiAABuanyBSM, PiAABuFC)

    def test_se_AA_tad(self):
        PiAATAanyBSM = sum(SMall.process('A','A',momenta=[0], only_topologies=["TwoPointTA"]).values())
        test.assert_allclose(0, PiAATAanyBSM, rtol = 0, atol = 1e-10)

    def test_dse_AA(self):
        dPiAAtotanyBSM = sum(SMall.process('A','A',momenta=[0], derivative = True).values())
        dPiAAtotFC = - alpha/(6*pi)*(loopfunctions.MSDR-4*loopfunctions.B0(0,Mt**2,Mt**2)+6*loopfunctions.B0(0,MWp**2,MWp**2)-16*loopfunctions.dB00(0,Mt**2,Mt**2)+18*loopfunctions.dB00(0,MWp**2,MWp**2))
        test.assert_allclose(dPiAAtotanyBSM,dPiAAtotFC)

    def test_se_AZ_A_scalar(self):
        """ test photon-Z self-energy """
        Zsign = SM.getSignSinThetaW()
        PiAZAsanyBSM = sum(SMscalar.process('A','Z',momenta=[0], only_topologies=["TwoPointA"]).values())
        PiAZAsFC = alpha/(4*pi*sqrt(CW2*SW2))*(CW2-SW2)*a0mw
        test.assert_allclose(Zsign*PiAZAsanyBSM,PiAZAsFC)

    def test_se_AZ_A_gauge(self):
        Zsign = SM.getSignSinThetaW()
        PiAZAvFC = - alpha/(2*pi*sqrt(SW2))*sqrt(CW2)*(2*loopfunctions.MSDR*MWp**2-3*a0mw)
        PiAZAvanyBSM = sum(SMgauge.process('A','Z',momenta=[0], only_topologies=["TwoPointA"]).values())
        test.assert_allclose(Zsign*PiAZAvanyBSM,PiAZAvFC)

    def test_se_AZ_B_scalar_gauge(self):
        Zsign = SM.getSignSinThetaW()
        PiAZBsvanyBSM = sum(SMboson.process('A','Z',momenta=[0], only_topologies=["TwoPointB"]).values())
        PiAZBsvFC = - alpha/(2*pi*sqrt(CW2))*MWp**2*sqrt(SW2)*loopfunctions.B0(0,MWp**2,MWp**2)
        PiAZBvFC = alpha/(2*pi)*sqrt(CW2/SW2)*(2*loopfunctions.MSDR*MWp**2-a0mw-MWp**2*loopfunctions.B0(0,MWp**2,MWp**2)-5*loopfunctions.B00(0,MWp**2,MWp**2))
        PiAZBsFC = - alpha/(2*pi*sqrt(CW2*SW2))*(CW2-SW2)*loopfunctions.B00(0,MWp**2,MWp**2)
        PiAZBuFC = alpha/(2*pi)*sqrt(CW2/SW2)*loopfunctions.B00(0,MWp**2,MWp**2)
        test.assert_allclose(Zsign*PiAZBsvanyBSM,PiAZBsFC+PiAZBvFC+PiAZBsvFC+PiAZBuFC)

    def test_se_AZ_B_top(self):
        Zsign = SM.getSignSinThetaW()
        PiAZBtopanyBSM = sum(SMtop.process('A','Z',momenta=[0], only_topologies=["TwoPointB"]).values())
        PiAZBtopFC = alpha/(6*pi*sqrt(SW2*CW2))*(-3+8*SW2)*(a0mt-2*loopfunctions.B00(0,Mt**2,Mt**2))
        test.assert_allclose(Zsign*PiAZBtopanyBSM,PiAZBtopFC)

    def test_se_AZ_B_ghost(self):
        Zsign = SM.getSignSinThetaW()
        PiAZBuanyBSM = sum(SMghost.process('A','Z',momenta=[0], only_topologies=["TwoPointB"]).values())
        PiAZBuFC = alpha/(2*pi)*sqrt(CW2/SW2)*loopfunctions.B00(0,MWp**2,MWp**2)
        test.assert_allclose(Zsign*PiAZBuanyBSM,PiAZBuFC)

    def test_se_AZ_tad(self):
        PiAZTAanyBSM = sum(SM.process('A','Z',momenta=[0], only_topologies=["TwoPointTA"]).values())
        test.assert_allclose(PiAZTAanyBSM,0, rtol = 0, atol = 1e-10)

    def test_lambdahhh_OS(self):
        """ test whether result for lambdahhh has changed w.r.t. previous versions"""
        SM.load_renormalization_scheme('OS')
        test.assert_allclose(SM.lambdahhh(momenta = ['auto'])['total'].real, 179.367762)

    def test_lambdahhh_MS(self):
        """ test whether result for lambdahhh has changed w.r.t. previous versions"""
        SM.load_renormalization_scheme('MS')
        test.assert_allclose(SM.lambdahhh(momenta = ['auto'])['total'].real, 167.036221)

    def test_MW(self):
        """ test whether result for MW has changed w.r.t. previous versions"""
        SM.load_renormalization_scheme('OS')
        test.assert_allclose(SM.MW(), 80.353908)



import logging
from tqdm.auto import tqdm
from numpy import pi, sqrt
from numpy.linalg import eig
from itertools import combinations_with_replacement
from anyBSM.anyModel import anyModel

logging.basicConfig(format='%(levelname)s:%(message)s')
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

__doc__ = """
<div class="mermaid">
flowchart RL
    subgraph classes
    anyModel  --> anyPerturbativeUnitarity;
    end
    subgraph methods
    anyPerturbativeUnitarity --> eigSSSS
    end
</div>
"""

class anyPerturbativeUnitarity(anyModel):
    r""" Calculation of unitarity constraints

    Example
    ```python
    from anyBSM import anyPerturbativeUnitarity
    SM = anyPerturbativeUnitarity('SM')
    SM.eigSSSS() # returns largest SS->SS eigenvalue
    ```
    """

    def eigSSSS(self, ignore_scalars: list = [], parameters: dict = {}) -> float:
        """
           Args:
                ignore_scalars: list of scalars to be excluded from the scatter matrix
                parameters: optionally run `anyBSM.anyModel.setparameters(parameters)` so set input parameters
           Return:
               largest scattering eigenvalue in the s->infinity limit
        """
        if self.evaluation != 'numerical':
            logger.error('Can only compute unitarity constraints in numerical evaluation mode at the moment. (run `set_evaluation_mode("numerical")` or `-e numerical`)')
            return
        if parameters:
            self.setparameters(params=parameters)
        scalars = [s for s in self.particles["S"] if s.color == 1 and s not in ignore_scalars and s.name not in ignore_scalars]
        scatteringpairs = set(combinations_with_replacement(scalars,2))
        logging.debug(f'Found {len(scatteringpairs)} scattering pairs.')
        smatrix = []
        for s1 in tqdm(scatteringpairs, desc = 'Calculating SS->SS scattering matrix in the high-energy limit',  disable = not self.progress, leave = False):
            row = []
            for s2 in tqdm(scatteringpairs, disable = not self.progress, leave = False):
                sym1 = sqrt(2) if s1[0] == s1[1] else 1
                sym2 = sqrt(2) if s2[0] == s2[1] else 1
                coup = self.getcoupling(s1[0],s1[1],s2[0],s2[1])
                coup = complex(0,-1)*coup['c'].nvalue/(sym1*sym2) if coup else 0.0
                coup = coup.real
                row.append(coup)
            smatrix.append(row)
        return max(abs(eig(smatrix)[0]))/(16*pi)



#!/usr/bin/env python3
import unittest
from anyBSM import anyBSM
import numpy.testing as test
import logging
logging.getLogger('anyBSM').setLevel(logging.WARNING)

def scaledep(model,states):
    lamQ1 = model.lambdahhh(fields=states, parameters = {'Qren': 172.5})['total'].real
    lamQ2 = model.lambdahhh(fields=states, parameters = {'Qren': 17250000})['total'].real
    return lamQ1, lamQ2

SM = anyBSM('SM', quiet = True, progress = False, scheme_name = 'OS')

THDM = anyBSM('THDMII', quiet = True, progress = False, caching = 2, scheme_name = 'OS')
THDM.setparameters({'alphaH': 0.1, 'Mh1':125.1,'Mh2':400,'MAh2':410,'MHm2': 420, 'TanBeta': 3,'M': 430})

class numericTest(unittest.TestCase):
    """ checks UV-finiteness of lam_ijk in various BSM models by checking the scale-dependence of OS schemes  """

    def test_UV_SM(self):
        """ test UV-finiteness in the SM for lamhhh """
        test.assert_allclose(*scaledep(SM,['h','h','h']))

    def test_UV_THDM111_no_alignment(self):
        """ test UV-finiteness in the THDM for lamhhh away from the alignment limit """
        test.assert_allclose(*scaledep(THDM,['h1','h1','h1']))

    def test_UV_THDM112_no_alignment(self):
        """ test UV-finiteness in the THDM for lamhhH away from the alignment limit """
        test.assert_allclose(*scaledep(THDM,['h1','h1','h2']))

    def test_UV_THDM122_no_alignment(self):
        """ test UV-finiteness in the THDM for lamhHH away from the alignment limit """
        test.assert_allclose(*scaledep(THDM,['h1','h2','h2']))

    def test_UV_THDM222_no_alignment(self):
        """ test UV-finiteness in the THDM for lamHHH away from the alignment limit """
        test.assert_allclose(*scaledep(THDM,['h2','h2','h2']))



#!/usr/bin/env python3

import unittest
import numpy.testing as test
from anyBSM import anyBSM

SM = anyBSM('SM', progress = False, quiet = True)
SM.clear_cache()
SM.set_evaluation_mode('numerical')

""" this plugs numbers directly into generic diagrams """
SM.caching = 0
lam_no_cache = SM.lambdahhh()['total'].real

""" this will first generate the abbreviated/analytical
result, store it in cache and then evaluate the cached
result numerically at each run. Thus, it is expected to
slightly differ from the `lam_no_cache` result."""
SM.caching = 2
lam_cache_1 = SM.lambdahhh()['total'].real
lam_cache_2 = SM.lambdahhh()['total'].real

class numericTest(unittest.TestCase):
    def test_caching_0(self):
        test.assert_allclose(lam_no_cache, lam_cache_1, rtol = 1e-10, atol = 0)

    def test_caching_1(self):
        test.assert_allclose(lam_no_cache, lam_cache_2, rtol = 1e-10, atol = 0)

    def test_caching_2(self):
        test.assert_allclose(lam_cache_1, lam_cache_2, rtol = 1e-15, atol = 0)



from __future__ import division
import logging
from anyBSM.anyModel import anyModel
from anyBSM.anyH3 import anyH3
from anyBSM.anyEWPO import anyEWPO
from anyBSM.anyPerturbativeUnitarity import anyPerturbativeUnitarity
from anyBSM.anyEffectiveCoupling import anyEffectiveCoupling
from anyBSM.anyHH import anyHH

logging.basicConfig(format='%(levelname)s:%(message)s')
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
__doc__ = """
<div class="mermaid">
flowchart RL

  subgraph user[main classes]
    direction TB
    anyModel --> anyProcess
    anyProcess --> anyH3
    anyProcess --> anyEWPO
    anyProcess --> ...
    anyH3 --> anyEffectiveCoupling
    anyH3 --> anyBSM
    ... --> anyBSM
  end
  subgraph internal[internal modules]
    direction TB
    A[topologies<br>diagrams<br>loopfunctions]
  end
  subgraph external[external optional modules]
    direction TB
    pyCollier
  end
  internal --- user
  external --- user
</div>"""
class anyBSM(anyHH, anyEffectiveCoupling, anyEWPO, anyH3, anyPerturbativeUnitarity):
    """ Wrapper class which inherits from all available classes that define physics obserables.
    Example:
    ```python
    from anyBSM import anyBSM
    SM = anyBSM('SM', scheme_name = 'OS')
    # all methods from all observables are available such as:
    SM.lambdahhh() # the trilinear coupling (from anyBSM.anyH3)
    SM.MW() # the W-mass (from anyBSM.anyEWPO)
    ```
    For more information on how two initialize a model see anyBSM.anyModel or consult the online documentation.
    """
    def __init__(self, *args, **kwargs):
        """ Takes the same arguments as anyBSM.anyModel.anyModel.__init__ """
        anyModel.__init__(self, *args, **kwargs)



#!/usr/bin/env python3
import unittest
from anyBSM import anyBSM
import numpy.testing as test
import logging
from math import pi
logging.getLogger('anyBSM').setLevel(logging.WARNING)

class numericTest(unittest.TestCase):
    """ checks equivalence between FJ and OS treatment of tadpoles in aligned SM and THDM """

    def test_SM(self):
        SM = anyBSM('SM', quiet = True, progress = False)
        SM.clear_cache()
        SM.load_renormalization_scheme('OSFJ')
        lam_FJ = SM.lambdahhh()['total']
        SM.load_renormalization_scheme('OS')
        lam_tOS = SM.lambdahhh()['total']
        test.assert_allclose(lam_FJ,lam_tOS)

    def test_THDM(self):
        THDM = anyBSM('THDMII', quiet = True, progress = False)
        THDM.clear_cache()
        THDM.setparameters({'alphaH': THDM.all_parameters['betaH']-pi/2.})
        THDM.load_renormalization_scheme('OSalignment')
        lam_FJ = THDM.lambdahhh()['total']
        THDM.load_renormalization_scheme('OSalignmentadpoles')
        lam_tOS = THDM.lambdahhh()['total']
        test.assert_allclose(lam_FJ,lam_tOS)



from anyBSM import anyBSM as anyBSM_single
from numpy import array
from collections import defaultdict
from typing import get_type_hints
from tqdm import tqdm
from multiprocessing import Process, Queue
import logging

__doc__ = """
# Basic Parallelism
Example usage:
```python
from anyBSM.parallel import anyBSM
import numpy as np

# initialize 8 parallel worker models and one master-model
SM=anyBSM('SM',num_workers=8);

# set variable for all workers
SM.to_workers('warnSSSS', False)

# read variable from all workers
SM.from_workers('warnSSSS')

# prepare (dict-typed) input
inputs = [{'parameters': {'Mh': mh}} for mh in np.linspace(20,1000,50)]

# compute lambdahhh for all inputs, distributed across all workers
results = SM.lambdahhh(worker_kwargs=inputs,transpose=True)

# compute max SS->SS eigenvalue for all inputs, distributed across all workers
results = SM.eigSSSS(worker_kwargs=inputs)

# prepare another (list-typed) input and compute differential HH production xs
inputs = [mh**2 for mh in np.linspace(250,1001,20)]
SM.sigmahat_ggHH(worker_args=inputs,pdf_scale=250,transpose=True)

```
"""

def list2dict(ll, fast = False):
    """ convert list-of-dicts to dict-of-lists """
    if fast:
        result = defaultdict(list)
        [ result[k].append(v) for k,v in ll.items() for entry in ll ]
        return result

    result = defaultdict(list)
    keys = set()
    for entry in ll:
        if type(entry) is dict:
            keys = keys.union(set(entry.keys()))
    for k in keys:
        result[k] = []
        for entry in ll:
            if type(entry) is dict and k in entry:
                result[k].append(entry[k])
            else:
                result[k].append(entry)
    return result

class anyBSMWorker(Process):
    """ anyBSM worker process
    Spawns an anyBSM instance with given setup_kwargs.
    Listens on task_queue and writes results into result_queue
    """
    def __init__(self,
                 task_queue: Queue,
                 result_queue: Queue,
                 model : str,
                 setup_kwargs: dict = {}):
        super().__init__()
        self.setup_kwargs = {'quiet' : True, 'progress': False, 'ask': False}
        self.setup_kwargs.update(setup_kwargs)

        self.task_queue = task_queue
        self.result_queue = result_queue

        self.model = model

    def run(self):
        model = anyBSM_single(self.model, **self.setup_kwargs)

        while True:
            job = self.task_queue.get()
            if job == "STOP":
                break
            if len(job) == 2:
                if job[0] == "SET":
                    try:
                        setattr(model, job[1][0], job[1][1])
                        self.result_queue.put((job[1][0], job[1][1]))
                    except Exception as e:
                        self.result_queue.put((job, str(e)))
                    continue
                if job[0] == "GET":
                    try:
                        self.result_queue.put((f'{self.name}: {job[1]} =', getattr(model,job[1])))
                    except Exception as e:
                        self.result_queue.put((job, str(e)))
                    continue

            if len(job) != 3:
                self.result_queue.put((job, 'ERROR: job requires exactly three arguments (id, func_name, (*args, **kwargs))'))
                continue

            if job[1] == "EVAL":
                try:
                    result = model._eval(job[1][0],**job[1][1])
                    self.result_queue.put((job[0], job[1][1], result))
                except Exception as e:
                    self.result_queue.put((job, str(e)))
                continue

            eval_func = job[1]
            if isinstance(job[2], dict):
                kwargs = job[2]
                args = []
            else:
                args, kwargs = job[2]
            try:
                method = getattr(model, eval_func)
                result = method(*args, **kwargs)
                # print(self.name, eval_func, args,kwargs)
                self.result_queue.put((job[0], result))
            except Exception as e:
                self.result_queue.put((job[0], str(e)))


class anyBSM(anyBSM_single):
    def __init__(self, model : str, num_workers : int = 4, **kwargs):
        super().__init__(model, **kwargs)
        self._master_manager  = True
        self._model = model
        self._num_workers = num_workers
        self._setup_kwargs = kwargs
        self._task_queues = []
        self._result_queue = Queue()
        self._workers = []

        self.respawn_workers()

    def respawn_workers(self):
        self._shutdown()
        for i in range(self._num_workers):
            tq = Queue()
            worker = anyBSMWorker(tq, self._result_queue, self._model, self._setup_kwargs)
            worker.start()
            self._task_queues.append(tq)
            self._workers.append(worker)

    # def __setattr__(self, name, value):
    #     super().__setattr__(name, value)

    #     # Skip local-only attributes or private ones
    #     if name.startswith('_') or not hasattr(self, '_num_workers'):
    #         return

    #     # Broadcast to all workers
    #     for tq in self._task_queues:
    #         tq.put(("SET",[name, value]))
    #     for _ in range(self._num_workers):
    #         self._result_queue.get()

    def __getattribute__(self, name):
        if name.startswith("_") or not hasattr(self, '_master_manager'):
            return object.__getattribute__(self, name)

        attr = super().__getattribute__(name)

        # Wrap callable methods for optional parallelism
        if callable(attr) and not name.startswith("_"):
            def wrapper(*args,worker_args=[], worker_kwargs=[], transpose=False, **kwargs):
                inputs = []
                if worker_args and worker_kwargs:
                    if len(worker_args) != len(worker_kwargs):
                        logging.error(f'worker_args and worker_kwargs should be of same length: {len(worker_args)}!={len(worker_kwargs)}')
                    inputs = [[a,{**k, **kwargs}] for a,k in zip(worker_args,worker_kwargs)]
                elif worker_args:
                    shape = array(worker_args).shape
                    if len(shape) == 1:
                        inputs = [[[a],kwargs] for a in worker_args]
                    else:
                        inputs = [[a,kwargs] for a in worker_args]
                elif worker_kwargs:
                    inputs = [[args,{**k, **kwargs}] for k in worker_kwargs]
                if inputs:
                    if args and not worker_kwargs:
                        logging.warning('additional arguments ignored')
                    return self.parallel_map(name, inputs, transpose)
                return attr(*args, **kwargs)
            signature = f'type hints of non-paralel function: {get_type_hints(attr)}'
            if attr.__doc__:
                wrapper.__doc__ = attr.__doc__ + signature
            return wrapper

        return attr

    def parallel_map(self, func_name, inputs,transpose = False):
        li = len(inputs)
        for i, item in enumerate(inputs):
            self._task_queues[i % self._num_workers].put((i, func_name, item))

        results = []
        for _ in tqdm(range(li), desc=f"Parallel {func_name}"):
            results.append(self._result_queue.get())

        results = dict(results)
        results = [results[i] for i in range(li)]
        if transpose:
            hints = get_type_hints(getattr(anyBSM_single,func_name))
            htype = hints.get('return',None)
            if htype is dict:
                results = list2dict(results)
            elif htype is tuple or htype is list:
                results = array(results).transpose()
        return results

    def parallel_eval(self, string, inputs):
        """ runs anyBSM._eval(string, **inputs) with inputs in parallel.
        Args:
           * string: a str to evaluate; any string compatible with the current model
           * inputs: a list of dictionaries. Entries are computed in parallel
        """
        li = len(inputs)
        for i, item in enumerate(inputs):
            self._task_queues[i % self._num_workers].put(('EVAL', (i, string, item)))
        results = []
        for _ in tqdm(range(li), desc="Evaluating"):
            results.append(self._result_queue.get())
        results = dict(results)
        return [results[i] for i in range(li)]

    def from_workers(self, attr):
        for tq in self._task_queues:
            tq.put(("GET", attr))

        return [self._result_queue.get() for _ in self._task_queues]

    def to_workers(self, name, value):
        for tq in self._task_queues:
            tq.put(("SET",[name, value]))
        for _ in range(self._num_workers):
            self._result_queue.get()

    def _shutdown(self):
        for tq in self._task_queues:
            tq.put("STOP")
        for w in self._workers:
            w.join()

    def __exit__(self):
        self._shutdown()

    def __del__(self):
        self._shutdown()

    def __enter__(self):
        return self



#!/usr/bin/env python3
import unittest
from numpy import arctan, arcsin
from anyBSM import anyPerturbativeUnitarity
import numpy.testing as test
import logging
logging.getLogger('anyBSM').setLevel(logging.WARNING)


THDM = anyPerturbativeUnitarity('THDMII', quiet = True, progress = False)

# result obtained with SPheno using the same default input parameters
eigSSSS_SPheno = 0.52212574

class numericTest(unittest.TestCase):

    def test_eigSSSS(self):
        """ compute and compare largest SS->SS scattering eigenvalue """
        eigSSSS_anyBSM = THDM.eigSSSS(parameters = {'alphaH': arctan(2)-arcsin(0.9), 'TanBeta':2, 'Mh1':125.1, 'Mh2': 300., 'MAh2': 300., 'MHm2': 550., 'M': 0})
        test.assert_allclose(eigSSSS_SPheno, eigSSSS_anyBSM)



import cmath

def complexconjugate(z):
    return z.conjugate()

conjugate = complexconjugate

Abs = abs

def Re(z):
    return z.real
re = Re

def Im(z):
    return z.imag
im = Im

I = complex(0,1)

# New functions (trigonometric)

def sec(z):
    return 1./cmath.cos(z.real)

def asec(z):
    return cmath.acos(1./(z.real))

def csc(z):
    return 1./cmath.sin(z.real)

def acsc(z):
    return cmath.asin(1./(z.real))

def cot(z):
    return 1./cmath.tan(z.real)

# Heaviside theta function

def theta_function(x,y,z):
    return y if x else z

# Auxiliary functions for NLO
def cond(cond, exprTrue, exprFalse):
    return exprTrue if cond == 0.0 else exprFalse

def reglog(z):
    return 0.0 if z == 0.0 else cmath.log(z.real)



#!/usr/bin/env python3

import unittest
import numpy.testing as test
from random import random
from anyBSM import loopfunctions
from itertools import combinations_with_replacement


B0 = loopfunctions.B0_p0 # our own analytical results
C0 = loopfunctions.C0_p0
B00 = loopfunctions.B00_p0
dB0 = loopfunctions.dB0_p0
dB00 = loopfunctions.dB00_p0
cB0 = loopfunctions.pyCollier.b0 # collier
cB00 = loopfunctions.pyCollier.b00 # collier
cC0 = loopfunctions.pyCollier.c0
cdB0 = loopfunctions.pyCollier.db0
cdB00 = loopfunctions.pyCollier.db00

loopfunctions.set_renscale(172.5)

x = random()*1000
y = random()*1000
z = random()*1000

class numericTest(unittest.TestCase):
    """ IR divergent cases are not tested since these are handled differently in COLLIER and in the analytical expresssions """

    # test B0 from COLLIER against analytical B0 implementation
    def test_B0_xy(self):
        test.assert_allclose(B0(x,y), cB0(0,x,y))

    def test_B0_xx(self):
        test.assert_allclose(B0(x,x), cB0(0,x,x))

    def test_B0_x0(self):
        test.assert_allclose(B0(x,0), cB0(0,x,0))

    def test_B0_0y(self):
        test.assert_allclose(B0(0,y), cB0(0,0,y))

    # test dB0 from COLLIER against analytical B0 implementation
    def test_dB0_xy(self):
        test.assert_allclose(dB0(x,y), cdB0(0,x,y))

    def test_dB0_xx(self):
        test.assert_allclose(dB0(x,x), cdB0(0,x,x))

    def test_dB0_x0(self):
        test.assert_allclose(dB0(x,0), cdB0(0,x,0))

    def test_dB0_0y(self):
        test.assert_allclose(dB0(0,y), cdB0(0,0,y))

    # test B00 from COLLIER against analytical B0 implementation
    def test_B00_xy(self):
        test.assert_allclose(B00(x,y), cB00(0,x,y))

    def test_B00_xx(self):
        test.assert_allclose(B00(x,x), cB00(0,x,x))

    def test_B00_x0(self):
        test.assert_allclose(B00(x,0), cB00(0,x,0))

    def test_B00_0y(self):
        test.assert_allclose(B00(0,y), cB00(0,0,y))

    # test dB00 from COLLIER against analytical B0 implementation
    def test_dB00_xy(self):
        test.assert_allclose(dB00(x,y), cdB00(0,x,y))

    def test_dB00_xx(self):
        test.assert_allclose(dB00(x,x), cdB00(0,x,x))

    def test_dB00_x0(self):
        test.assert_allclose(dB00(x,0), cdB00(0,x,0))

    def test_dB00_0y(self):
        test.assert_allclose(dB00(0,y), cdB00(0,0,y))

    # test C0 from COLLIER against analytical C0 implementation
    def test_C0(self):
        for p in combinations_with_replacement([x,y,z,0],3):
            if p[1] != 0 and p[2] != 0:
                test.assert_allclose(C0(*p), cC0(0,0,0,*p))



import re
from sympy.printing.latex import LatexPrinter
from anyBSM.utils import fieldtypes

function_names = {
        'A0': 'A_0',
        'B0': 'B_0',
        'dB0': r'\dot{B}_0',
        'B00': 'B_{00}',
        'dB00': r'\dot{B}_{00}',
        'C0': 'C_0',
        'C1': 'C_1',
        'C2': 'C_2',
        'lnbar': r'\overline{\ln}',
        }

class anyBSMPrinter(LatexPrinter):
    """ Custom LaTeX printer: defines how certain sympy object are represented in LaTeX output """
    def _print_conjugate(self, expr, exp=None):
        tex = r"{{%s}^*}" % self._print(expr.args[0])

        if exp is not None:
            return r"%s^{%s}" % (tex, exp)
        else:
            return tex

    def _print_Function(self, expr, exp=None):
        func = expr.func.__name__
        if func in function_names.keys():
            args = ','.join([self._print(i) for i in expr.args])
            return f'{function_names[func]}({args})'
        return super()._print_Function(expr, exp=exp)

def latex(expr,**settings):
    """ Executor for custom LaTeX printer `anyBSMPrinter()` """
    return anyBSMPrinter(settings).doprint(expr)

def particle_name(particle):
    """ Tries to convert a particle name into valid LaTeX.
    Takes an `anyBSM.ufo.object_library.Particle()` or a string as input.
    If the input is already latex, it is returned without modification.
    """
    if not isinstance(particle, str) and particle.__class__.__name__ == 'Particle':
        texname = particle.texname
    else:
        texname = particle
    if '\\' in texname or '^' in texname or '_' in texname: # is most likely already tex
        return texname
    if len(texname) == 1:
        return texname
    if 'bar' in texname or '~' in texname:
        name = particle_name(texname.replace('bar','').replace('~',''))
        return f'\\tilde{{{name}}}'
    if particle.__class__.__name__ == 'Particle':
        if texname.endswith('p') or texname.endswith('+') and particle.charge > 0:
            name = particle_name(texname[:-1])
            return f'{{{name}}}^+'
        if texname.endswith('m') or texname.endswith('-') and particle.charge < 0:
            name = particle_name(texname[:-1])
            return f'{{{name}}}^-'
    texname = re.sub(r'^([a-zA-Z])(\d)$', r'\1_{\2}', texname)
    texname = re.sub(r'^([a-zA-Z])([A-Za-z])$', r'\1_{\2}', texname)
    return texname

def parameter_name(param):
    """ Same as `particle_name()` but for generic couplings and model parameters """
    texname = str(param)
    if '\\' in texname or '^' in texname or '{' in texname: # is most likely already tex
        return texname
    texname = re.sub(r'GC_(\d+)', r'G_C^{\1}', texname)
    texname = re.sub(r'^([a-zA-Z])_(\d+)$', r'\1_{\2}', texname)
    texname = re.sub(r'^([a-zA-Z])([a-zA-Z])_(\d+)$', r'\1_{\2}^{\3}', texname)
    texname = re.sub(r'^(\w)(\d+)$', r'\1_{\2}', texname)
    return texname

def line(particle):
    """ Guess the propagator line of a `anyBSM.ufo.object_library.Particle()` based on its spin and charge """
    ftype = {-1 : 'ghost', 1: 'scalar', 2: 'fermion', 3: 'boson'}.get(particle.spin, None)
    if not ftype:
        return 'plain'
    if particle.charge != 0 and ftype in ['scalar', 'boson']:
        ftype = 'charged ' + ftype
    if particle.selfconjugate and ftype == 'fermion':
        ftype = 'majorana'
    if particle.pdg_code < 0 and ftype != 'ghost':
        ftype = 'anti ' + ftype
    return ftype

def texname(particle, generic=False, external=False):
    """ Either return particles TeX-name or the generic field type """
    if not generic or external:
        return particle_name(particle.texname)
    ftype = fieldtypes([particle])
    return ftype



#!/bin/python
from __future__ import division
from collections import defaultdict
import logging

logger = logging.getLogger('anyBSM.topologies')

__doc__ = """ # AnyTopology
A module defining all generic topologies.\n
A topology consists of an ordered list of vertices.
Each vertex consists of an ordered list of fields.
Each field is a generic object that can be an internal/external/loop-field as well as a anti-field.\n

The pre-defined topologies are stored in anyBSM.topologies.TOPOLOGIES:
```yml
TOPOLOGIES = {
    <number of external fields> : {
        <topology name> : <topology object>,
        ...
        }
        ...
    }
```
e.g. TOPOLOGIES[2]['TwoPointB'] contains the B0-type self-energy
topology.

The topologies are populated with fields of a given UFO model in
`anyBSM.anyBSM.insert_fields()`.
"""

TOPOLOGIES = defaultdict(dict)
""" Global variable used to store topologies """


class Field():
    """ Representation of generic field """
    def __init__(self, label, insertion = None, anti = False):
        self.label = label
        self.insertion = None
        self.antifield = anti

class InternalField(Field):
    pass

class LoopField(Field):
    pass

class ExternalField(Field):
    pass

class Vertex():
    """ Representation of generic vertex """
    def __init__(self, *fields):
        self.fields = list(fields)

    def setinsertion(self, fields):
        for f in self.fields:
            fieldtoinsert = fields.get(f.label, None)
            if not fieldtoinsert:
                continue
            if fieldtoinsert.__class__.__name__ != 'Particle':
                logger.error('Can only insert `Particle` instances into Topology(), not {} ({})'.format(fieldtoinsert, fieldtoinsert.__class__.__name__))
                return
            f.insertion = fieldtoinsert
            if f.antifield:
                f.insertion = f.insertion.anti()

    def getinsertion(self):
        return [f.insertion for f in self.fields]

class VertexCT(Vertex):
    pass

def isexternal(field):
    return isinstance(field, ExternalField)
def isinternal(field):
    return type(field) in [InternalField, LoopField]
def isloop(field):
    return isinstance(field, LoopField)


class Topology():
    """ Representation of generic one-loop topology """
    def __init__(self, name, *vertices, symmetryfactor = 1, tadpole = False, wfr = False):
        self.vertices = vertices
        self.extfields = {} # OrderedDict()
        self.intfields = {} # OrderedDict()
        self.loopfields = {} # OrderedDict()
        self.Amplitude = name
        self.symmetryfactor = symmetryfactor
        self.tadpole = tadpole
        self.wfr = wfr
        allfields = [field for v in vertices for field in v.fields]
        for field in allfields:
            if isexternal(field):
                self.extfields.update({field.label: field})
            if isinternal(field):
                self.intfields.update({field.label: field})
                if any([ f.label == field.label and f.antifield != field.antifield for f in allfields]):
                    self.loopfields.update({field.label: field})

        self.next = len(self.extfields)
        self.nint = len(self.intfields)
        self.nloop = len(self.loopfields)

        TOPOLOGIES[self.next].update({name : self})

    def getinsertion(self):
        return [v.getinsertion() for v in self.vertices]

    def insertexternal(self, *fields):
        assert len(fields) == self.next
        for v in self.vertices:
            v.setinsertion({label:field for label,field in zip(self.extfields.keys(), fields)})

    def insertinternal(self, *fields):
        assert len(fields) == self.nint
        for v in self.vertices:
            v.setinsertion({label:field for label,field in zip(self.intfields.keys(), fields)})

"""
Define all possible one-loop topologies
"""
# 1-point
Topology(
        'OnePointA',
        Vertex(ExternalField(0),LoopField(1),LoopField(1,anti=True)),
)
# 1-point CT-inserted diagram:
# Tadpoles always MSbar -> no need for tadpole-CT diagrams, but have them here in comments
# Topology(
#         'OnePointCT',
#         VertexCT(LoopField(0)),
# )
# 2-point
Topology(
        'TwoPointA',
        Vertex(ExternalField(0), LoopField(1), LoopField(1,anti=True), ExternalField(2)),
)
Topology(
        'TwoPointB',
        Vertex(ExternalField(0),LoopField(1),LoopField(2,anti=True)),
        Vertex(ExternalField(3,anti=True), LoopField(1),LoopField(2,anti=True)),
)
Topology(
        'TwoPointTA', # tadpole contribution to selfenergy
        Vertex(ExternalField(0),InternalField(1,anti=True),ExternalField(2)),
        Vertex(InternalField(1),LoopField(3,anti=True),LoopField(3)),
        tadpole = 1
)
# 2-point CT-inserted diagrams
# Topology( # mass CT contribution
#         'TwoPointCT',
#         Vertex(ExternalField(0),ExternalField(1,anti=True)),
# )
# Topology(
#         'TwoPointTCT', # tadpole CT contribution
#         Vertex(ExternalField(0),InternalField(1),ExternalField(2)),
#         VertexCT(InternalField(1)),
#         tadpole = 1
# )

# 3-point
Topology(
        'ThreePointTree',
        Vertex(ExternalField(0),ExternalField(1),ExternalField(2)),
)
Topology(
        'ThreePointB', # bubble diagram
        Vertex(ExternalField(0),LoopField(3,anti=True),LoopField(4)),
        Vertex(ExternalField(1),ExternalField(2),LoopField(4,anti=True),LoopField(3)),
        symmetryfactor = 3
)
Topology(
        'ThreePointC', # triangle diagram
        Vertex(ExternalField(0),LoopField(3),LoopField(5, anti=True)),
        Vertex(ExternalField(1),LoopField(4),LoopField(3, anti=True)),
        Vertex(ExternalField(2),LoopField(5),LoopField(4, anti=True)),
)
Topology(
        'ThreePointTA', # 4-point vertex where one leg is a attached to a tadpole
        Vertex(ExternalField(0),InternalField(1,anti=True),ExternalField(2),ExternalField(3)),
        Vertex(InternalField(1),LoopField(4,anti=True),LoopField(4)),
        tadpole = 1
)
Topology(
        'ThreePointWFRA', # external leg with A-bubble
        Vertex(ExternalField(0),InternalField(1,anti=True),ExternalField(2)),
        Vertex(InternalField(1),LoopField(3,anti=True),LoopField(3),ExternalField(4)),
        symmetryfactor = 3,
        wfr = True
)
Topology(
        'ThreePointWFRB', # external leg with B-bubble
        Vertex(ExternalField(0),InternalField(1,anti=True),ExternalField(2)),
        Vertex(InternalField(1,anti=True),LoopField(3),LoopField(4,anti=True)),
        Vertex(ExternalField(5),LoopField(3),LoopField(4,anti=True)),
        symmetryfactor = 3,
        wfr = True
)
Topology(
        'ThreePointWFRT', # external leg with Tadpole
        Vertex(ExternalField(0),ExternalField(1),InternalField(3,anti=True)),
        Vertex(ExternalField(2),InternalField(3),InternalField(4)),
        Vertex(InternalField(4),LoopField(5),LoopField(5,anti=True)),
        symmetryfactor = 3,
        wfr = True,
        tadpole = 4
)
# 3-point CT-inserted diagrams
# Topology(
#         'ThreePointCT', # Vertex CT
#         VertexCT(ExternalField(0),ExternalField(1),ExternalField(2)),
# )
# Topology(
#         'ThreePointMCT', # mass CT on external leg
#         Vertex(ExternalField(0),InternalField(1,anti=True),ExternalField(2)),
#         VertexCT(InternalField(1),ExternalField(3)),
#         wfr = True
# )
# Topology(
#         'ThreePointTCT', # tadpole CT on 4-vertex leg
#         Vertex(ExternalField(0),InternalField(1,anti=True),ExternalField(2),ExternalField(3)),
#         VertexCT(InternalField(1)),
#         tadpole = 1
# )
# Topology(
#         'ThreePointWTCT', # external leg with tadpole CT
#         Vertex(ExternalField(0),InternalField(1,anti=True),ExternalField(2)),
#         Vertex(InternalField(1),InternalField(3, anti=True),ExternalField(4)),
#         VertexCT(InternalField(3)),
#         wfr = True,
#         tadpole = 1
# )



from __future__ import division
import logging
from cmath import pi, sqrt
from anyBSM.Delta_r_SM import Delta_r_SM
import anyBSM.physic_utils as physics_utils
from anyBSM.anyProcess import anyProcess

logging.basicConfig(format='%(levelname)s:%(message)s')
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

__doc__ = """
#  Electroweak precision observables (EWPOs)
<div class="mermaid">
flowchart RL
    subgraph classes
    anyModel --> anyProcess --> anyEWPO
    end
    subgraph methods/EWPOs
    anyEWPO --> Delta_r --> M_W
    anyEWPO --> ...
    end
</div>
"""


class anyEWPO(anyProcess):
    """ A class which collects calculations for electroweak precision observables.
    This class inherits from `anyBSM.anyProcess` which itself inherits from `anyBSM.anyModel`.
    For more information on how to initialize this class, see the documentation of the mother-class `anyBSM.anyModel.anyModel.__init__` or the online documentation.

    The class `anyBSM.anyEWPO` currently implements two methods:
      * `anyBSM.anyEWPO.anyEWPO.Delta_r()`: caclulates $\\Delta r$
      * `anyBSM.anyEWPO.anyEWPO.MW()`: calculates $M_W$

    It is assumed that a proper renormalization scheme has already been loaded and all SM-like particles/parameters are correctly identified.
    To ensure this `anyBSM.anyModel.anyModel.load_renormalization_scheme()` should be issued beforehand (see example below).

    Example:
    ```python
    from anyBSM import anyBSM
    SSM = anyBSM('SSM', scheme_name = 'OS')
    SSM.MW()
    ```
    which calculates $M_W$ in the singlet extended SM at the one-loop order including
    all known higher-order SM-corrections to $\\Delta r$ (see `anyBSM.Delta_r_SM`).
    """

    def Delta_r(self, MWrun: float, MWfw: float, MZrun: float, MZfw: float, incl_SM_ho: bool = True, **kwargs) -> float:
        """ Calculates $\\Delta r$.

        Args:
            MWrun: $M_W$ defined using a running width
            MWfw: $M_W$ defined using a fixed width
            MZrun: $M_Z$ defined using a running width
            MZfw:  $M_Z$ defined using a fixed width
            incl_SM_ho: whether to include higher-order SM corrections
            kwargs: all arguments supported by anyBSM.process
            parameters: optionally run `anyBSM.anyModel.setparameters(parameters)` so set input parameters

        Returns:
            $\\Delta r$
        """

        # set parameters
        MWfw2 = f'({MWfw}**2)'
        MZfw2 = f'({MZfw}**2)'
        CW = self._eval(f'({MWfw}/{MZfw})')
        SW = self._eval(f'sqrt(1-{MWfw2}/{MZfw2})')
        aEW = self._eval(f"1/({self.SM_parameters['alphaQEDinverse'].name})")
        dalpha = self._eval(self.SM_parameters['Dalpha'].name)
        GFermi = self._eval(self.SM_parameters['GFermi'].name)
        asMZ = self._eval(f"{self.SM_parameters['alphaQCD'].name}")

        ZA = f"(SignSinThetaW*({self.Sigma(self.all_particles['Z-Boson'], self.all_particles['Photon'], momentum=0, **kwargs)})/{MZfw2})"

        deltaEL = f"1/2*({dalpha} +\
                ({self.Sigma(self.all_particles['Photon'], momentum = 'MZ**2', exclude_particles_pdg=self.heavy_fields, **kwargs)})/{MZfw2}\
                + {self.Sigmaprime(self.all_particles['Photon'], exclude_particles_pdg=self.light_SM_fields, **kwargs)})\
                + {SW}/({CW})*{ZA}"

        Wselfp0 = f"(({self.Sigma(self.all_particles['W-Boson'], momentum=0, **kwargs)})/{MWfw2})"

        WselfOS = f"(({self.Sigma(self.all_particles['W-Boson'], **kwargs)})/{MWfw2})"
        ZselfOS = f"(({self.Sigma(self.all_particles['Z-Boson'], **kwargs)})/{MZfw2})"
        dSW = f"-{SW}/2*{CW}**2/{SW}**2*({WselfOS} - {ZselfOS})"

        # TODO: implement generic {F,F}->{F,F} at zero external momenta as well as generic fermion selfenergies
        VertexBoxZf = f"-2/({CW}*{SW})*{ZA} \
                + ({aEW})/(4*pi*{SW}**2)*(6+(7-4*{SW}**2)/(2*{SW}**2)*log({CW}**2))"

        dr = self._eval(f"{Wselfp0}-{WselfOS} \
                + 2*({deltaEL}) - 2*({dSW})/{SW} + {VertexBoxZf}")

        if self.evaluation != 'numerical':
            if incl_SM_ho:
                logger.warning("Can't include higher-order SM corrections to $Delta r$ in analytical mode.")
            return dr

        # calculate SM higher-order corrections
        if incl_SM_ho:
            MH = self.getmass('Higgs-Boson')
            MT = self.getmass('Top-Quark')
            MB = self.getmass('Bottom-Quark')
            MC = self.getmass('Charm-Quark')
            MS = self.getmass('Strange-Quark')
            MU = self.getmass('Up-Quark')
            MD = self.getmass('Down-Quark')
            ML = self.getmass('Tau-Lepton')
            MM = self.getmass('Muon-Lepton')
            ME = self.getmass('Electron-Lepton')
            asMT = physics_utils.RunAlfas(MZfw, MT, asMZ, looplevel=3)

            _, drSM2Lrem, drSMho = Delta_r_SM(MWrun, MWfw, MZrun, MZfw, MH, MT, MB, aEW, asMT, dalpha, GFermi, ME, MM, ML, MU, MD, MC, MS)

            dr += dalpha**2 + 2*dalpha*(dr - dalpha) + drSM2Lrem + drSMho  # dalpha is contained in dr1L (see also footnote 6 of 1506.07465)

        return dr.real

    def MW(self, MWstart: float = 80.0, precision: int = 8, incl_SM_ho: bool = True, parameters: dict = {}, **kwargs) -> float:
        """ Iteratively calculates $M_W$ from $G_{Fermi}$ and corrections to the muon decay
        (i.e. $\\Delta r$, see anyEWPO.Delta_r).

        Args:
            MWstart: start value of the iteration
            precision: stop iteration if last `precision`-digits do
                not change w.r.t the previous iteration
            incl_SM_ho: whether to include the higher-order
                corrections from the Standard Model (with the one-loop
                result being subtracted)
            kwargs: all arguments supported by anyBSM.process

        Returns:
            $M_W$
        """
        if self.evaluation != 'numerical':
            logger.error('Can only compute MW iteratively in numerical evaluation mode. (run `set_evaluation_mode("numerical")` or `-e numerical`)')
            return
        if parameters:
            self.setparameters(params=parameters)
        if not self.SM_parameters['alphaQEDinverse']:
            self.find_SM_parameters()
        if not self.SM_particles['W-Boson']:
            self.find_SM_particles()
        GFermi = self.SM_parameters['GFermi'].value
        MWsave = self.SM_particles['W-Boson'].nmass
        MWpara = self.SM_particles['W-Boson'].mass.name
        MZrun = self.SM_particles['Z-Boson'].nmass
        MZpara = self.SM_particles['Z-Boson'].mass.name
        aEW = 1/self.SM_parameters['alphaQEDinverse'].value
        asMZ = self._eval(f"{self.SM_parameters['alphaQCD'].name}")
        asMW = physics_utils.RunAlfas(MZrun, MWsave, asMZ,looplevel=3)
        cnt = 0
        MWa = MWstart
        MWb = -12345.0

        # ensure that one-loop corrections to Delta_r use fixed-width Z boson mass if higher order corrections are activated
        if incl_SM_ho:
            MZfw = physics_utils.MZfw(MZrun)
            self.setparameters({MZpara: MZfw})
        else:
            MZfw = MZrun

        while abs(MWa - MWb) > 10**(-precision):
            if cnt == 0:
                MWb = MWa
            cnt += 1
            if cnt > 99:
                logger.error(f'MW iteration did not converge in-time.\
                        Last two values: MW = {MWa} GeV and MW = {MWb} GeV.')
                return
            self.setparameters({MWpara: MWb})
            MWrun = physics_utils.MWrun(MWb, asMW, GFermi)
            Deltar = self.Delta_r(MWrun, MWb, MZrun, MZfw, incl_SM_ho=incl_SM_ho, **kwargs)
            MWa = MWb
            MWb = sqrt(MZfw**2*(1/2+sqrt(1/4-pi*aEW/(sqrt(2)*GFermi*MZfw**2)*(1 + Deltar))))
            if abs(MWb.imag) > 1e-6:
                logger.warning(f'MW developed imaginary part: {MWb} GeV')
            logger.info(f'MW iteration {cnt}: MW = {MWb} GeV; Delta_r = {Deltar}.')
            MWb = MWb.real
        if incl_SM_ho:
            MWres = physics_utils.MWrun(MWb, asMW, GFermi)
        else:
            MWres = MWb

        # reset UFO parameter value
        self.setparameters({MWpara: MWsave, MZpara: MZrun})

        return MWres.real
